"""Snapshot tests for all backend combinations.

Uses external_file() from inline-snapshot to store generated code in files.
Snapshots are stored in tests/snapshots/ as importable Python packages.

Run with --inline-snapshot=create to generate initial snapshots.
Run with --inline-snapshot=fix to update snapshots.
"""

from __future__ import annotations

from pathlib import Path
from typing import Any

import pytest
from inline_snapshot import external_file

from tests.conftest import format_python

from senzo import generate_tree
from senzo.backends import HttpxBackend, MsgspecBackend, PydanticBackend
from senzo.backends.http.aiohttp import AiohttpBackend
from senzo.backends.http.requests import RequestsBackend


HTTPBIN_BACKEND_MATRIX = [
    pytest.param(
        "msgspec_httpx_sync",
        MsgspecBackend(),
        HttpxBackend(async_mode=False),
        id="msgspec_httpx_sync",
    ),
    pytest.param(
        "msgspec_httpx_async",
        MsgspecBackend(),
        HttpxBackend(async_mode=True),
        id="msgspec_httpx_async",
    ),
    pytest.param(
        "msgspec_requests",
        MsgspecBackend(),
        RequestsBackend(),
        id="msgspec_requests",
    ),
    pytest.param(
        "msgspec_aiohttp",
        MsgspecBackend(),
        AiohttpBackend(),
        id="msgspec_aiohttp",
    ),
    pytest.param(
        "pydantic_httpx_sync",
        PydanticBackend(),
        HttpxBackend(async_mode=False),
        id="pydantic_httpx_sync",
    ),
    pytest.param(
        "pydantic_httpx_async",
        PydanticBackend(),
        HttpxBackend(async_mode=True),
        id="pydantic_httpx_async",
    ),
    pytest.param(
        "pydantic_requests",
        PydanticBackend(),
        RequestsBackend(),
        id="pydantic_requests",
    ),
    pytest.param(
        "pydantic_aiohttp",
        PydanticBackend(),
        AiohttpBackend(),
        id="pydantic_aiohttp",
    ),
]


@pytest.mark.parametrize("name,dataclass_backend,http_backend", HTTPBIN_BACKEND_MATRIX)
def test_httpbin_client_snapshot(
    name: str,
    dataclass_backend: MsgspecBackend | PydanticBackend,
    http_backend: HttpxBackend | RequestsBackend | AiohttpBackend,
    httpbin_spec: dict[str, Any],
) -> None:
    """Test generated client matches snapshot for each backend combo."""
    package_name = f"httpbin_{name}"
    snapshot_dir = Path("snapshots") / package_name

    tree = generate_tree(
        httpbin_spec,
        package_name=package_name,
        dataclass_backend=dataclass_backend,
        http_backend=http_backend,
    )

    assert format_python(tree["client.py"].code) == external_file(
        snapshot_dir / "client.py"
    )
    assert format_python(tree["_base.py"].code) == external_file(
        snapshot_dir / "_base.py"
    )
    assert format_python(tree["types/models.py"].code) == external_file(
        snapshot_dir / "types" / "models.py"
    )
    assert format_python(tree["types/__init__.py"].code) == external_file(
        snapshot_dir / "types" / "__init__.py"
    )
    assert format_python(tree["__init__.py"].code) == external_file(
        snapshot_dir / "__init__.py"
    )


@pytest.mark.parametrize("name,dataclass_backend,http_backend", HTTPBIN_BACKEND_MATRIX)
def test_allof_client_snapshot(
    name: str,
    dataclass_backend: MsgspecBackend | PydanticBackend,
    http_backend: HttpxBackend | RequestsBackend | AiohttpBackend,
    allof_spec: dict[str, Any],
) -> None:
    """Test allOf schema handling for each backend combo."""
    package_name = f"allof_{name}"
    snapshot_dir = Path("snapshots") / package_name

    tree = generate_tree(
        allof_spec,
        package_name=package_name,
        dataclass_backend=dataclass_backend,
        http_backend=http_backend,
    )

    assert format_python(tree["types/models.py"].code) == external_file(
        snapshot_dir / "types" / "models.py"
    )


EDGE_CASES_BACKEND_MATRIX = [
    pytest.param(
        "msgspec_httpx",
        MsgspecBackend(),
        HttpxBackend(async_mode=False),
        id="msgspec_httpx",
    ),
    pytest.param(
        "pydantic_httpx",
        PydanticBackend(),
        HttpxBackend(async_mode=False),
        id="pydantic_httpx",
    ),
]


@pytest.mark.parametrize(
    "name,dataclass_backend,http_backend", EDGE_CASES_BACKEND_MATRIX
)
def test_edge_cases_snapshot(
    name: str,
    dataclass_backend: MsgspecBackend | PydanticBackend,
    http_backend: HttpxBackend,
    edge_cases_spec: dict[str, Any],
) -> None:
    """Test edge case handling (naming, enums, formats) for each dataclass backend."""
    package_name = f"edge_cases_{name}"
    snapshot_dir = Path("snapshots") / package_name

    tree = generate_tree(
        edge_cases_spec,
        package_name=package_name,
        dataclass_backend=dataclass_backend,
        http_backend=http_backend,
    )

    assert format_python(tree["types/models.py"].code) == external_file(
        snapshot_dir / "types" / "models.py"
    )
    assert format_python(tree["client.py"].code) == external_file(
        snapshot_dir / "client.py"
    )


SNAPSHOT_NAMES = [b.values[0] for b in HTTPBIN_BACKEND_MATRIX]  # type: ignore[attr-defined]


@pytest.mark.parametrize("name", SNAPSHOT_NAMES)
def test_httpbin_snapshot_importable(name: str) -> None:
    """Verify that each httpbin snapshot is a valid, importable Python package."""
    import importlib
    import sys

    snapshot_dir = Path(__file__).parent / "snapshots" / f"httpbin_{name}"
    if not snapshot_dir.exists():
        pytest.skip(f"Snapshot not found: {snapshot_dir}")

    if str(snapshot_dir.parent) not in sys.path:
        sys.path.insert(0, str(snapshot_dir.parent))

    try:
        module = importlib.import_module(f"httpbin_{name}")
        assert hasattr(module, "Client"), f"Missing Client class in httpbin_{name}"
        assert hasattr(module, "APIError"), f"Missing APIError class in httpbin_{name}"
    finally:
        if f"httpbin_{name}" in sys.modules:
            del sys.modules[f"httpbin_{name}"]


KALSHI_BACKEND_MATRIX = [
    pytest.param(
        "msgspec_httpx_sync",
        MsgspecBackend(rename=None),
        HttpxBackend(async_mode=False),
        id="msgspec_httpx_sync",
    ),
    pytest.param(
        "msgspec_httpx_async",
        MsgspecBackend(rename=None),
        HttpxBackend(async_mode=True),
        id="msgspec_httpx_async",
    ),
    pytest.param(
        "msgspec_requests",
        MsgspecBackend(rename=None),
        RequestsBackend(),
        id="msgspec_requests",
    ),
    pytest.param(
        "msgspec_aiohttp",
        MsgspecBackend(rename=None),
        AiohttpBackend(),
        id="msgspec_aiohttp",
    ),
    pytest.param(
        "pydantic_httpx_sync",
        PydanticBackend(),
        HttpxBackend(async_mode=False),
        id="pydantic_httpx_sync",
    ),
    pytest.param(
        "pydantic_httpx_async",
        PydanticBackend(),
        HttpxBackend(async_mode=True),
        id="pydantic_httpx_async",
    ),
    pytest.param(
        "pydantic_requests",
        PydanticBackend(),
        RequestsBackend(),
        id="pydantic_requests",
    ),
    pytest.param(
        "pydantic_aiohttp",
        PydanticBackend(),
        AiohttpBackend(),
        id="pydantic_aiohttp",
    ),
]


@pytest.mark.parametrize("name,dataclass_backend,http_backend", KALSHI_BACKEND_MATRIX)
def test_kalshi_client_snapshot(
    name: str,
    dataclass_backend: MsgspecBackend | PydanticBackend,
    http_backend: HttpxBackend | RequestsBackend | AiohttpBackend,
    kalshi_spec: dict[str, Any],
) -> None:
    """Test generated Kalshi client matches snapshot for each backend combo."""
    from senzo.settings import TagStyle

    package_name = f"kalshi_{name}"
    snapshot_dir = Path("snapshots") / package_name

    tree = generate_tree(
        kalshi_spec,
        package_name=package_name,
        dataclass_backend=dataclass_backend,
        http_backend=http_backend,
        tag_style=TagStyle.GROUPED,
    )

    assert format_python(tree["client.py"].code) == external_file(
        snapshot_dir / "client.py"
    )
    assert format_python(tree["_base.py"].code) == external_file(
        snapshot_dir / "_base.py"
    )
    assert format_python(tree["types/models.py"].code) == external_file(
        snapshot_dir / "types" / "models.py"
    )
    assert format_python(tree["types/__init__.py"].code) == external_file(
        snapshot_dir / "types" / "__init__.py"
    )
    assert format_python(tree["__init__.py"].code) == external_file(
        snapshot_dir / "__init__.py"
    )


@pytest.mark.parametrize("name", SNAPSHOT_NAMES)
def test_kalshi_snapshot_importable(name: str) -> None:
    """Verify that each Kalshi snapshot is a valid, importable Python package."""
    import importlib
    import sys

    snapshot_dir = Path(__file__).parent / "snapshots" / f"kalshi_{name}"
    if not snapshot_dir.exists():
        pytest.skip(f"Snapshot not found: {snapshot_dir}")

    if str(snapshot_dir.parent) not in sys.path:
        sys.path.insert(0, str(snapshot_dir.parent))

    try:
        module = importlib.import_module(f"kalshi_{name}")
        assert hasattr(module, "Client"), f"Missing Client class in kalshi_{name}"
        assert hasattr(module, "APIError"), f"Missing APIError class in kalshi_{name}"
    finally:
        if f"kalshi_{name}" in sys.modules:
            del sys.modules[f"kalshi_{name}"]


PAGINATION_BACKEND_MATRIX = [
    pytest.param(
        "msgspec_httpx_sync",
        MsgspecBackend(),
        HttpxBackend(async_mode=False),
        id="msgspec_httpx_sync",
    ),
    pytest.param(
        "msgspec_httpx_async",
        MsgspecBackend(),
        HttpxBackend(async_mode=True),
        id="msgspec_httpx_async",
    ),
]


@pytest.mark.parametrize(
    "name,dataclass_backend,http_backend", PAGINATION_BACKEND_MATRIX
)
def test_pagination_client_snapshot(
    name: str,
    dataclass_backend: MsgspecBackend | PydanticBackend,
    http_backend: HttpxBackend | RequestsBackend | AiohttpBackend,
    pagination_spec: dict[str, Any],
) -> None:
    """Test generated pagination client matches snapshot."""
    package_name = f"pagination_{name}"
    snapshot_dir = Path("snapshots") / package_name

    tree = generate_tree(
        pagination_spec,
        package_name=package_name,
        dataclass_backend=dataclass_backend,
        http_backend=http_backend,
        pagination={
            "listUsers": {
                "items": "items",
                "next_token": "cursor",
                "token_param": "cursor",
                "limit_param": "limit",
            },
        },
    )

    assert format_python(tree["client.py"].code) == external_file(
        snapshot_dir / "client.py"
    )
    assert format_python(tree["_base.py"].code) == external_file(
        snapshot_dir / "_base.py"
    )
    assert format_python(tree["types/models.py"].code) == external_file(
        snapshot_dir / "types" / "models.py"
    )
    assert format_python(tree["types/__init__.py"].code) == external_file(
        snapshot_dir / "types" / "__init__.py"
    )
    assert format_python(tree["__init__.py"].code) == external_file(
        snapshot_dir / "__init__.py"
    )


KALSHI_PAGINATION_BACKEND_MATRIX = [
    pytest.param(
        "msgspec_httpx_sync",
        MsgspecBackend(rename=None),
        HttpxBackend(async_mode=False),
        id="msgspec_httpx_sync",
    ),
    pytest.param(
        "msgspec_httpx_async",
        MsgspecBackend(rename=None),
        HttpxBackend(async_mode=True),
        id="msgspec_httpx_async",
    ),
]


@pytest.mark.parametrize(
    "name,dataclass_backend,http_backend", KALSHI_PAGINATION_BACKEND_MATRIX
)
def test_kalshi_pagination_client_snapshot(
    name: str,
    dataclass_backend: MsgspecBackend | PydanticBackend,
    http_backend: HttpxBackend | RequestsBackend | AiohttpBackend,
    kalshi_spec: dict[str, Any],
) -> None:
    """Test generated Kalshi client with pagination matches snapshot."""
    from senzo.settings import TagStyle

    package_name = f"kalshi_paginated_{name}"
    snapshot_dir = Path("snapshots") / package_name

    tree = generate_tree(
        kalshi_spec,
        package_name=package_name,
        dataclass_backend=dataclass_backend,
        http_backend=http_backend,
        tag_style=TagStyle.GROUPED,
        pagination={
            "GetOrders": {
                "items": "orders",
                "next_token": "cursor",
                "token_param": "cursor",
                "limit_param": "limit",
            },
            "GetFills": {
                "items": "fills",
                "next_token": "cursor",
                "token_param": "cursor",
                "limit_param": "limit",
            },
            "GetTrades": {
                "items": "trades",
                "next_token": "cursor",
                "token_param": "cursor",
                "limit_param": "limit",
            },
        },
    )

    assert format_python(tree["client.py"].code) == external_file(
        snapshot_dir / "client.py"
    )
    assert format_python(tree["_base.py"].code) == external_file(
        snapshot_dir / "_base.py"
    )
    assert format_python(tree["types/models.py"].code) == external_file(
        snapshot_dir / "types" / "models.py"
    )
    assert format_python(tree["types/__init__.py"].code) == external_file(
        snapshot_dir / "types" / "__init__.py"
    )
    assert format_python(tree["__init__.py"].code) == external_file(
        snapshot_dir / "__init__.py"
    )



from __future__ import annotations
import msgspec
from typing import final


@final
class Animal(msgspec.Struct, frozen=True, rename="camel"):
    """Base animal type"""

    id: str
    name: str


@final
class Cat(msgspec.Struct, frozen=True, rename="camel"):
    """Base animal type"""

    id: str
    name: str
    indoor: bool | None = None


@final
class Dog(msgspec.Struct, frozen=True, rename="camel"):
    """Base animal type"""

    breed: str
    id: str
    name: str
    age: int | None = None


@final
class Labrador(msgspec.Struct, frozen=True, rename="camel"):
    """A labrador retriever"""

    breed: str
    color: str
    id: str
    name: str
    age: int | None = None



from __future__ import annotations
import msgspec
from typing import final
from collections.abc import Sequence


@final
class GetItemsResponse(msgspec.Struct, frozen=True, rename="camel"):
    data: Sequence[Item]
    next_cursor: str | None = None


@final
class Item(msgspec.Struct, frozen=True, rename="camel"):
    id: int
    title: str


@final
class ListOrdersResponse(msgspec.Struct, frozen=True, rename="camel"):
    orders: Sequence[Order]


@final
class ListUsersResponse(msgspec.Struct, frozen=True, rename="camel"):
    cursor: str
    items: Sequence[User]


@final
class Order(msgspec.Struct, frozen=True, rename="camel"):
    amount: float
    id: str


@final
class User(msgspec.Struct, frozen=True, rename="camel"):
    email: str
    id: str
    name: str



from __future__ import annotations
import msgspec
from enum import Enum
import datetime
import uuid
from typing import final, Literal


@final
class FormatTypes(msgspec.Struct, frozen=True, rename="camel"):
    """Tests various format types"""

    created_date: datetime.date | None = None
    created_datetime: datetime.datetime | None = None
    item_uuid: uuid.UUID | None = None


@final
class HttpRequest(msgspec.Struct, frozen=True, rename="camel"):
    """Tests snake_case conversion of HTTPRequest -> http_request"""

    created_at: datetime.datetime | None = None
    http_header: str | None = None


@final
class ItemWithInlineEnum(msgspec.Struct, frozen=True, rename="camel"):
    """Tests inline enum without explicit type"""

    kind: Literal["a", "b", "c"] | None = None


class NullableStatus(str, Enum):
    ON = "on"
    OFF = "off"


class Status(str, Enum):
    ACTIVE = "active"
    IN_PROGRESS = "in-progress"
    _100_ = "100%"
    COMPLETED = "COMPLETED"



from __future__ import annotations
import msgspec
from typing import final, Any
from collections.abc import Mapping


@final
class BearerResponse(msgspec.Struct, frozen=True, rename="camel"):
    authenticated: bool
    token: str


@final
class HeadersResponse(msgspec.Struct, frozen=True, rename="camel"):
    headers: Mapping[str, str]


@final
class HttpMethodResponse(msgspec.Struct, frozen=True, rename="camel"):
    args: Mapping[str, str]
    headers: Mapping[str, str]
    origin: str
    url: str


@final
class IpResponse(msgspec.Struct, frozen=True, rename="camel"):
    origin: str


@final
class PostResponse(msgspec.Struct, frozen=True, rename="camel"):
    args: Mapping[str, str]
    headers: Mapping[str, str]
    origin: str
    url: str
    data: str | None = None
    json: Mapping[str, Any] | None = None


@final
class UserAgentResponse(msgspec.Struct, frozen=True, rename="camel"):
    user_agent: str



from __future__ import annotations
from typing import final
import httpx
from typing import Any
from ._base import APIError, ResponseValue


@final
class Client:
    def __init__(self, base_url: str) -> None:
        self._base_url = base_url
        self._client = httpx.Client(base_url=base_url)

    def close(self) -> None:
        self._client.close()

    def __enter__(self) -> "Client":
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        self.close()

    def list_items(
        self, limit: int | None = None, http_header: str | None = None
    ) -> ResponseValue[None]:
        _path = "/items"
        _params: dict[str, Any] = {"limit": limit, "HTTPHeader": http_header}
        _url = self._base_url.rstrip("/") + _path
        _response = self._client.request(
            method="GET",
            url=_url,
            headers={},
            params={k: v for k, v in _params.items() if v is not None},
            content=None,
        )
        _headers_dict = dict(_response.headers)
        if _response.status_code >= 400:
            raise APIError(
                status_code=_response.status_code,
                body=_response.content,
                headers=_headers_dict,
            )
        return ResponseValue(
            value=None, status_code=_response.status_code, headers=_headers_dict
        )



from .models import (
    AcceptQuoteRequest,
    AmendOrderRequest,
    AmendOrderResponse,
    Announcement,
    ApiKey,
    ApplySubaccountTransferRequest,
    ApplySubaccountTransferResponse,
    AssociatedEvent,
    BatchCancelOrdersIndividualResponse,
    BatchCancelOrdersRequest,
    BatchCancelOrdersResponse,
    BatchCreateOrdersIndividualResponse,
    BatchCreateOrdersRequest,
    BatchCreateOrdersResponse,
    BatchGetMarketCandlesticksResponse,
    BidAskDistribution,
    CancelOrderResponse,
    CreateApiKeyRequest,
    CreateApiKeyResponse,
    CreateMarketInMultivariateEventCollectionRequest,
    CreateMarketInMultivariateEventCollectionResponse,
    CreateOrderGroupRequest,
    CreateOrderGroupResponse,
    CreateOrderRequest,
    CreateOrderResponse,
    CreateQuoteRequest,
    CreateQuoteResponse,
    CreateRfqRequest,
    CreateRfqResponse,
    CreateSubaccountResponse,
    DailySchedule,
    DecreaseOrderRequest,
    DecreaseOrderResponse,
    DepthQuery,
    EmptyResponse,
    ErrorResponse,
    EventData,
    EventPosition,
    ExchangeInstance,
    ExchangeStatus,
    Fill,
    FixedPointCount,
    FixedPointDollars,
    ForecastPercentilesPoint,
    GenerateApiKeyRequest,
    GenerateApiKeyResponse,
    GetApiKeysResponse,
    GetBalanceResponse,
    GetCommunicationsIdResponse,
    GetEventCandlesticksResponse,
    GetEventForecastPercentilesHistoryResponse,
    GetEventMetadataResponse,
    GetEventResponse,
    GetEventsResponse,
    GetExchangeAnnouncementsResponse,
    GetExchangeScheduleResponse,
    GetFillsResponse,
    GetFiltersBySportsResponse,
    GetIncentiveProgramsResponse,
    GetLiveDataResponse,
    GetLiveDatasResponse,
    GetMarketCandlesticksResponse,
    GetMarketOrderbookResponse,
    GetMarketResponse,
    GetMarketsResponse,
    GetMilestoneResponse,
    GetMilestonesResponse,
    GetMultivariateEventCollectionLookupHistoryResponse,
    GetMultivariateEventCollectionResponse,
    GetMultivariateEventCollectionsResponse,
    GetMultivariateEventsResponse,
    GetOrderGroupResponse,
    GetOrderGroupsResponse,
    GetOrderQueuePositionResponse,
    GetOrderQueuePositionsResponse,
    GetOrderResponse,
    GetOrdersResponse,
    GetPortfolioRestingOrderTotalValueResponse,
    GetPositionsResponse,
    GetQuoteResponse,
    GetQuotesResponse,
    GetRfQsResponse,
    GetRfqResponse,
    GetSeriesFeeChangesResponse,
    GetSeriesListResponse,
    GetSeriesResponse,
    GetSettlementsResponse,
    GetStructuredTargetResponse,
    GetStructuredTargetsResponse,
    GetSubaccountBalancesResponse,
    GetSubaccountTransfersResponse,
    GetTagsForSeriesCategoriesResponse,
    GetTradesResponse,
    GetUserDataTimestampResponse,
    IncentiveProgram,
    IntraExchangeInstanceTransferRequest,
    IntraExchangeInstanceTransferResponse,
    LiveData,
    LookupPoint,
    LookupTickersForMarketInMultivariateEventCollectionRequest,
    LookupTickersForMarketInMultivariateEventCollectionResponse,
    MaintenanceWindow,
    Market,
    MarketCandlestick,
    MarketCandlesticksResponse,
    MarketMetadata,
    MarketPosition,
    Milestone,
    MultivariateEventCollection,
    MveSelectedLeg,
    Order,
    OrderGroup,
    OrderQueuePosition,
    OrderStatus,
    Orderbook,
    OrderbookCountFp,
    OrderbookLevel,
    PercentilePoint,
    PriceDistribution,
    PriceLevel,
    PriceLevelDollars,
    PriceLevelDollarsCountFp,
    PriceRange,
    Quote,
    Rfq,
    Schedule,
    ScopeList,
    SelfTradePreventionType,
    Series,
    SeriesFeeChange,
    Settlement,
    SettlementSource,
    SportFilterDetails,
    StructuredTarget,
    SubaccountBalance,
    SubaccountTransfer,
    TickerPair,
    Trade,
    WeeklySchedule,
)

__all__ = [
    "AcceptQuoteRequest",
    "AmendOrderRequest",
    "AmendOrderResponse",
    "Announcement",
    "ApiKey",
    "ApplySubaccountTransferRequest",
    "ApplySubaccountTransferResponse",
    "AssociatedEvent",
    "BatchCancelOrdersIndividualResponse",
    "BatchCancelOrdersRequest",
    "BatchCancelOrdersResponse",
    "BatchCreateOrdersIndividualResponse",
    "BatchCreateOrdersRequest",
    "BatchCreateOrdersResponse",
    "BatchGetMarketCandlesticksResponse",
    "BidAskDistribution",
    "CancelOrderResponse",
    "CreateApiKeyRequest",
    "CreateApiKeyResponse",
    "CreateMarketInMultivariateEventCollectionRequest",
    "CreateMarketInMultivariateEventCollectionResponse",
    "CreateOrderGroupRequest",
    "CreateOrderGroupResponse",
    "CreateOrderRequest",
    "CreateOrderResponse",
    "CreateQuoteRequest",
    "CreateQuoteResponse",
    "CreateRfqRequest",
    "CreateRfqResponse",
    "CreateSubaccountResponse",
    "DailySchedule",
    "DecreaseOrderRequest",
    "DecreaseOrderResponse",
    "DepthQuery",
    "EmptyResponse",
    "ErrorResponse",
    "EventData",
    "EventPosition",
    "ExchangeInstance",
    "ExchangeStatus",
    "Fill",
    "FixedPointCount",
    "FixedPointDollars",
    "ForecastPercentilesPoint",
    "GenerateApiKeyRequest",
    "GenerateApiKeyResponse",
    "GetApiKeysResponse",
    "GetBalanceResponse",
    "GetCommunicationsIdResponse",
    "GetEventCandlesticksResponse",
    "GetEventForecastPercentilesHistoryResponse",
    "GetEventMetadataResponse",
    "GetEventResponse",
    "GetEventsResponse",
    "GetExchangeAnnouncementsResponse",
    "GetExchangeScheduleResponse",
    "GetFillsResponse",
    "GetFiltersBySportsResponse",
    "GetIncentiveProgramsResponse",
    "GetLiveDataResponse",
    "GetLiveDatasResponse",
    "GetMarketCandlesticksResponse",
    "GetMarketOrderbookResponse",
    "GetMarketResponse",
    "GetMarketsResponse",
    "GetMilestoneResponse",
    "GetMilestonesResponse",
    "GetMultivariateEventCollectionLookupHistoryResponse",
    "GetMultivariateEventCollectionResponse",
    "GetMultivariateEventCollectionsResponse",
    "GetMultivariateEventsResponse",
    "GetOrderGroupResponse",
    "GetOrderGroupsResponse",
    "GetOrderQueuePositionResponse",
    "GetOrderQueuePositionsResponse",
    "GetOrderResponse",
    "GetOrdersResponse",
    "GetPortfolioRestingOrderTotalValueResponse",
    "GetPositionsResponse",
    "GetQuoteResponse",
    "GetQuotesResponse",
    "GetRfQsResponse",
    "GetRfqResponse",
    "GetSeriesFeeChangesResponse",
    "GetSeriesListResponse",
    "GetSeriesResponse",
    "GetSettlementsResponse",
    "GetStructuredTargetResponse",
    "GetStructuredTargetsResponse",
    "GetSubaccountBalancesResponse",
    "GetSubaccountTransfersResponse",
    "GetTagsForSeriesCategoriesResponse",
    "GetTradesResponse",
    "GetUserDataTimestampResponse",
    "IncentiveProgram",
    "IntraExchangeInstanceTransferRequest",
    "IntraExchangeInstanceTransferResponse",
    "LiveData",
    "LookupPoint",
    "LookupTickersForMarketInMultivariateEventCollectionRequest",
    "LookupTickersForMarketInMultivariateEventCollectionResponse",
    "MaintenanceWindow",
    "Market",
    "MarketCandlestick",
    "MarketCandlesticksResponse",
    "MarketMetadata",
    "MarketPosition",
    "Milestone",
    "MultivariateEventCollection",
    "MveSelectedLeg",
    "Order",
    "OrderGroup",
    "OrderQueuePosition",
    "OrderStatus",
    "Orderbook",
    "OrderbookCountFp",
    "OrderbookLevel",
    "PercentilePoint",
    "PriceDistribution",
    "PriceLevel",
    "PriceLevelDollars",
    "PriceLevelDollarsCountFp",
    "PriceRange",
    "Quote",
    "Rfq",
    "Schedule",
    "ScopeList",
    "SelfTradePreventionType",
    "Series",
    "SeriesFeeChange",
    "Settlement",
    "SettlementSource",
    "SportFilterDetails",
    "StructuredTarget",
    "SubaccountBalance",
    "SubaccountTransfer",
    "TickerPair",
    "Trade",
    "WeeklySchedule",
]



"""Live integration tests against Kalshi demo API.

These tests hit the real Kalshi demo API and require credentials:
- KALSHI_API_KEY_ID: Your API key ID
- KALSHI_PRIVATE_KEY: Path to PEM file or PEM string

Run with: uv run pytest -m live
Skip with: uv run pytest -m "not live"
"""

from __future__ import annotations

import base64
import importlib
import os
import sys
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import pytest
import yaml

from senzo.backends.dataclass.msgspec import MsgspecBackend
from senzo.backends.dataclass.pydantic import PydanticBackend
from senzo.backends.http.aiohttp import AiohttpBackend
from senzo.backends.http.httpx import HttpxBackend
from senzo.backends.http.requests import RequestsBackend
from senzo.generator import Generator, write_package
from senzo.parser import OpenAPISpec
from senzo.settings import HooksConfig, PaginationConfig, TagStyle
from tests.kalshi_auth import KalshiAuthState, sign_request, sign_request_sync

# Demo environment base URL
KALSHI_DEMO_URL = "https://demo-api.kalshi.co/trade-api/v2/"

# Backend matrix for live tests
LIVE_BACKEND_MATRIX = [
    pytest.param("msgspec_httpx_sync", id="msgspec_httpx_sync"),
    pytest.param("msgspec_httpx_async", id="msgspec_httpx_async"),
    pytest.param("msgspec_requests", id="msgspec_requests"),
    pytest.param("msgspec_aiohttp", id="msgspec_aiohttp"),
    pytest.param("pydantic_httpx_sync", id="pydantic_httpx_sync"),
    pytest.param("pydantic_httpx_async", id="pydantic_httpx_async"),
    pytest.param("pydantic_requests", id="pydantic_requests"),
    pytest.param("pydantic_aiohttp", id="pydantic_aiohttp"),
]


def _is_async_backend(name: str) -> bool:
    """Check if backend is async based on name."""
    return "async" in name or "aiohttp" in name


def _get_snapshot_module(name: str) -> Any:
    """Import the generated snapshot module for a backend."""
    snapshot_dir = Path(__file__).parent / "snapshots"
    if str(snapshot_dir) not in sys.path:
        sys.path.insert(0, str(snapshot_dir))
    return importlib.import_module(f"kalshi_{name}")


# Cache for generated auth clients to avoid regenerating on each test
_auth_client_cache: dict[str, tuple[Any, Path]] = {}
_auth_paginated_client_cache: dict[str, tuple[Any, Path]] = {}
_auth_temp_dir: tempfile.TemporaryDirectory[str] | None = None


def _get_auth_client_module(backend_name: str) -> Any:
    """Generate a Kalshi client with auth hooks baked in.

    Generates the client once and caches it for reuse across tests.
    """
    global _auth_temp_dir

    if backend_name in _auth_client_cache:
        return _auth_client_cache[backend_name][0]

    # Create temp dir if needed
    if _auth_temp_dir is None:
        _auth_temp_dir = tempfile.TemporaryDirectory()
        if _auth_temp_dir.name not in sys.path:
            sys.path.insert(0, _auth_temp_dir.name)

    # Load the Kalshi OpenAPI spec
    spec_path = Path(__file__).parent / "fixtures" / "kalshi.yaml"
    with open(spec_path) as f:
        spec_dict = yaml.safe_load(f)
    spec = OpenAPISpec.from_dict(spec_dict)

    # Determine backends based on name
    is_async = _is_async_backend(backend_name)
    is_pydantic = "pydantic" in backend_name

    # Use rename=None for msgspec to match Kalshi's snake_case JSON response
    dataclass_backend = (
        PydanticBackend() if is_pydantic else MsgspecBackend(rename=None)
    )

    if "aiohttp" in backend_name:
        http_backend = AiohttpBackend()
    elif "requests" in backend_name:
        http_backend = RequestsBackend()
    else:
        http_backend = HttpxBackend(async_mode=is_async)

    # Configure hooks
    hooks = HooksConfig(
        inner_type=KalshiAuthState,
        pre_hook=sign_request if is_async else sign_request_sync,
    )

    # Generate client
    package_name = f"kalshi_auth_{backend_name}"
    generator = Generator(
        spec=spec,
        package_name=package_name,
        dataclass_backend=dataclass_backend,
        http_backend=http_backend,
        tag_style=TagStyle.GROUPED,
        hooks=hooks,
    )

    modules = generator.generate()

    # Write to temp directory
    assert _auth_temp_dir is not None
    output_dir = Path(_auth_temp_dir.name) / package_name
    write_package(modules, str(output_dir))

    # Import the module
    module = importlib.import_module(package_name)
    _auth_client_cache[backend_name] = (module, output_dir)

    return module


@dataclass
class KalshiAuth:
    """Kalshi RSA-PSS authentication."""

    api_key_id: str
    private_key: Any  # RSAPrivateKey

    @classmethod
    def from_env(cls) -> KalshiAuth | None:
        """Load credentials from environment, return None if not set."""
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import serialization
        from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

        key_id = os.environ.get("KALSHI_API_KEY_ID")
        # Support both KALSHI_PRIVATE_KEY and KALSHI_API_PRIVATE_KEY
        private_key_value = os.environ.get("KALSHI_PRIVATE_KEY") or os.environ.get(
            "KALSHI_API_PRIVATE_KEY"
        )

        if not key_id or not private_key_value:
            return None

        # Load from file if it's a path
        if not private_key_value.startswith("-----BEGIN"):
            with open(private_key_value, "rb") as f:
                private_key = serialization.load_pem_private_key(
                    f.read(), password=None, backend=default_backend()
                )
        else:
            private_key = serialization.load_pem_private_key(
                private_key_value.encode(), password=None, backend=default_backend()
            )

        if not isinstance(private_key, RSAPrivateKey):
            raise TypeError("Key must be an RSA private key")

        return cls(api_key_id=key_id, private_key=private_key)

    def sign(self, message: str) -> str:
        """Sign a message with RSA-PSS."""
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric import padding

        signature = self.private_key.sign(
            message.encode("utf-8"),
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH,
            ),
            hashes.SHA256(),
        )
        return base64.b64encode(signature).decode("utf-8")


@pytest.fixture
def kalshi_auth() -> KalshiAuth | None:
    """Get Kalshi auth credentials if available."""
    return KalshiAuth.from_env()


@pytest.fixture
def require_kalshi_auth(kalshi_auth: KalshiAuth | None) -> KalshiAuth:
    """Require Kalshi auth credentials, skip if not available."""
    if kalshi_auth is None:
        pytest.skip("KALSHI_API_KEY_ID and KALSHI_PRIVATE_KEY not set")
    return kalshi_auth


# =============================================================================
# Public endpoint tests (no auth required)
# =============================================================================


@pytest.mark.live
@pytest.mark.parametrize("backend_name", LIVE_BACKEND_MATRIX)
async def test_get_exchange_status(backend_name: str) -> None:
    """Test fetching exchange status (public endpoint)."""
    module = _get_snapshot_module(backend_name)
    snapshot_dir = Path(__file__).parent / "snapshots" / f"kalshi_{backend_name}"
    if not snapshot_dir.exists():
        pytest.skip(f"Snapshot not generated: kalshi_{backend_name}")

    Client = module.Client

    if _is_async_backend(backend_name):
        async with Client(base_url=KALSHI_DEMO_URL) as client:
            response = await client.exchange.get_exchange_status()
            assert hasattr(response, "value")
            assert hasattr(response.value, "exchange_active")
    else:
        with Client(base_url=KALSHI_DEMO_URL) as client:
            response = client.exchange.get_exchange_status()
            assert hasattr(response, "value")
            assert hasattr(response.value, "exchange_active")


@pytest.mark.live
@pytest.mark.parametrize("backend_name", LIVE_BACKEND_MATRIX)
async def test_get_markets(backend_name: str) -> None:
    """Test fetching markets list (public endpoint)."""
    module = _get_snapshot_module(backend_name)
    snapshot_dir = Path(__file__).parent / "snapshots" / f"kalshi_{backend_name}"
    if not snapshot_dir.exists():
        pytest.skip(f"Snapshot not generated: kalshi_{backend_name}")

    Client = module.Client

    if _is_async_backend(backend_name):
        async with Client(base_url=KALSHI_DEMO_URL) as client:
            response = await client.market.get_markets(limit=5)
            assert hasattr(response, "value")
            assert hasattr(response.value, "markets")
            assert isinstance(response.value.markets, list)
    else:
        with Client(base_url=KALSHI_DEMO_URL) as client:
            response = client.market.get_markets(limit=5)
            assert hasattr(response, "value")
            assert hasattr(response.value, "markets")
            assert isinstance(response.value.markets, list)


# =============================================================================
# Authenticated endpoint tests
# =============================================================================


@pytest.mark.live
@pytest.mark.parametrize("backend_name", LIVE_BACKEND_MATRIX)
async def test_get_balance(backend_name: str, require_kalshi_auth: KalshiAuth) -> None:
    """Test fetching account balance (authenticated endpoint)."""
    module = _get_auth_client_module(backend_name)
    Client = module.Client

    # Create auth state from credentials
    auth_state = KalshiAuthState(
        api_key_id=require_kalshi_auth.api_key_id,
        private_key=require_kalshi_auth.private_key,
    )

    if _is_async_backend(backend_name):
        async with Client(base_url=KALSHI_DEMO_URL, inner=auth_state) as client:
            response = await client.portfolio.get_balance()
            assert hasattr(response, "value")
            assert hasattr(response.value, "balance")
            assert isinstance(response.value.balance, int)
    else:
        with Client(base_url=KALSHI_DEMO_URL, inner=auth_state) as client:
            response = client.portfolio.get_balance()
            assert hasattr(response, "value")
            assert hasattr(response.value, "balance")
            assert isinstance(response.value.balance, int)


@pytest.mark.live
@pytest.mark.parametrize("backend_name", LIVE_BACKEND_MATRIX)
async def test_get_positions(
    backend_name: str, require_kalshi_auth: KalshiAuth
) -> None:
    """Test fetching portfolio positions (authenticated endpoint)."""
    module = _get_auth_client_module(backend_name)
    Client = module.Client

    # Create auth state from credentials
    auth_state = KalshiAuthState(
        api_key_id=require_kalshi_auth.api_key_id,
        private_key=require_kalshi_auth.private_key,
    )

    if _is_async_backend(backend_name):
        async with Client(base_url=KALSHI_DEMO_URL, inner=auth_state) as client:
            response = await client.portfolio.get_positions()
            assert hasattr(response, "value")
            assert hasattr(response.value, "market_positions")
    else:
        with Client(base_url=KALSHI_DEMO_URL, inner=auth_state) as client:
            response = client.portfolio.get_positions()
            assert hasattr(response, "value")
            assert hasattr(response.value, "market_positions")


# =============================================================================
# Pagination tests (using paginated snapshots)
# =============================================================================

PAGINATION_BACKEND_MATRIX = [
    pytest.param("msgspec_httpx_sync", id="msgspec_httpx_sync"),
    pytest.param("msgspec_httpx_async", id="msgspec_httpx_async"),
]


def _get_paginated_snapshot_module(name: str) -> Any:
    """Import the generated paginated snapshot module for a backend."""
    snapshot_dir = Path(__file__).parent / "snapshots"
    if str(snapshot_dir) not in sys.path:
        sys.path.insert(0, str(snapshot_dir))
    return importlib.import_module(f"kalshi_paginated_{name}")


def _get_auth_paginated_client_module(backend_name: str) -> Any:
    """Generate a Kalshi client with auth hooks AND pagination baked in.

    Generates the client once and caches it for reuse across tests.
    """
    global _auth_temp_dir

    if backend_name in _auth_paginated_client_cache:
        return _auth_paginated_client_cache[backend_name][0]

    # Create temp dir if needed
    if _auth_temp_dir is None:
        _auth_temp_dir = tempfile.TemporaryDirectory()
        if _auth_temp_dir.name not in sys.path:
            sys.path.insert(0, _auth_temp_dir.name)

    # Load the Kalshi OpenAPI spec
    spec_path = Path(__file__).parent / "fixtures" / "kalshi.yaml"
    with open(spec_path) as f:
        spec_dict = yaml.safe_load(f)
    spec = OpenAPISpec.from_dict(spec_dict)

    # Determine backends based on name
    is_async = _is_async_backend(backend_name)

    # Use rename=None for msgspec to match Kalshi's snake_case JSON response
    dataclass_backend = MsgspecBackend(rename=None)
    http_backend = HttpxBackend(async_mode=is_async)

    # Configure hooks
    hooks = HooksConfig(
        inner_type=KalshiAuthState,
        pre_hook=sign_request if is_async else sign_request_sync,
    )

    # Pagination config for Kalshi endpoints
    pagination_config: dict[str, PaginationConfig] = {
        "GetOrders": PaginationConfig(
            items="orders",
            next_token="cursor",
            token_param="cursor",
            limit_param="limit",
        ),
        "GetFills": PaginationConfig(
            items="fills",
            next_token="cursor",
            token_param="cursor",
            limit_param="limit",
        ),
        "GetTrades": PaginationConfig(
            items="trades",
            next_token="cursor",
            token_param="cursor",
            limit_param="limit",
        ),
    }

    # Generate client
    package_name = f"kalshi_auth_paginated_{backend_name}"
    generator = Generator(
        spec=spec,
        package_name=package_name,
        dataclass_backend=dataclass_backend,
        http_backend=http_backend,
        tag_style=TagStyle.GROUPED,
        hooks=hooks,
        pagination=pagination_config,
    )

    modules = generator.generate()

    # Write to temp directory
    assert _auth_temp_dir is not None
    output_dir = Path(_auth_temp_dir.name) / package_name
    write_package(modules, str(output_dir))

    # Import the module
    module = importlib.import_module(package_name)
    _auth_paginated_client_cache[backend_name] = (module, output_dir)

    return module


@pytest.mark.live
@pytest.mark.parametrize("backend_name", PAGINATION_BACKEND_MATRIX)
async def test_get_trades_iter(backend_name: str) -> None:
    """Test paginated trades iteration (public endpoint)."""
    module = _get_paginated_snapshot_module(backend_name)
    snapshot_dir = (
        Path(__file__).parent / "snapshots" / f"kalshi_paginated_{backend_name}"
    )
    if not snapshot_dir.exists():
        pytest.skip(f"Snapshot not generated: kalshi_paginated_{backend_name}")

    Client = module.Client

    if _is_async_backend(backend_name):
        async with Client(base_url=KALSHI_DEMO_URL) as client:
            # Test that iterator works and returns items
            trades: list[Any] = []
            async for trade in client.market.get_trades_iter(limit=10):
                trades.append(trade)
                if len(trades) >= 5:
                    break

            assert len(trades) <= 5
            if trades:
                # Verify trade objects have expected attributes
                assert hasattr(trades[0], "trade_id")
                assert hasattr(trades[0], "ticker")
    else:
        with Client(base_url=KALSHI_DEMO_URL) as client:
            # Test that iterator works and returns items
            trades: list[Any] = []
            for trade in client.market.get_trades_iter(limit=10):
                trades.append(trade)
                if len(trades) >= 5:
                    break

            assert len(trades) <= 5
            if trades:
                # Verify trade objects have expected attributes
                assert hasattr(trades[0], "trade_id")
                assert hasattr(trades[0], "ticker")


@pytest.mark.live
@pytest.mark.parametrize("backend_name", PAGINATION_BACKEND_MATRIX)
async def test_get_trades_iter_take(backend_name: str) -> None:
    """Test paginated trades with take() method (public endpoint)."""
    module = _get_paginated_snapshot_module(backend_name)
    snapshot_dir = (
        Path(__file__).parent / "snapshots" / f"kalshi_paginated_{backend_name}"
    )
    if not snapshot_dir.exists():
        pytest.skip(f"Snapshot not generated: kalshi_paginated_{backend_name}")

    Client = module.Client

    if _is_async_backend(backend_name):
        async with Client(base_url=KALSHI_DEMO_URL) as client:
            # Test atake() method
            trades = await client.market.get_trades_iter(limit=10).atake(3)
            assert len(trades) <= 3
            if trades:
                assert hasattr(trades[0], "trade_id")
    else:
        with Client(base_url=KALSHI_DEMO_URL) as client:
            # Test take() method
            trades = client.market.get_trades_iter(limit=10).take(3)
            assert len(trades) <= 3
            if trades:
                assert hasattr(trades[0], "trade_id")


@pytest.mark.live
@pytest.mark.parametrize("backend_name", PAGINATION_BACKEND_MATRIX)
async def test_get_orders_iter(
    backend_name: str, require_kalshi_auth: KalshiAuth
) -> None:
    """Test paginated orders iteration (authenticated endpoint)."""
    module = _get_auth_paginated_client_module(backend_name)
    Client = module.Client

    # Create auth state from credentials
    auth_state = KalshiAuthState(
        api_key_id=require_kalshi_auth.api_key_id,
        private_key=require_kalshi_auth.private_key,
    )

    if _is_async_backend(backend_name):
        async with Client(base_url=KALSHI_DEMO_URL, inner=auth_state) as client:
            # Test that iterator works (may be empty if no orders)
            orders = await client.orders.get_orders_iter(limit=10).atake(5)
            # Just verify it returns a list (may be empty)
            assert isinstance(orders, list)
    else:
        with Client(base_url=KALSHI_DEMO_URL, inner=auth_state) as client:
            # Test that iterator works (may be empty if no orders)
            orders = client.orders.get_orders_iter(limit=10).take(5)
            # Just verify it returns a list (may be empty)
            assert isinstance(orders, list)


@pytest.mark.live
@pytest.mark.parametrize("backend_name", PAGINATION_BACKEND_MATRIX)
async def test_get_fills_iter(
    backend_name: str, require_kalshi_auth: KalshiAuth
) -> None:
    """Test paginated fills iteration (authenticated endpoint)."""
    module = _get_auth_paginated_client_module(backend_name)
    Client = module.Client

    # Create auth state from credentials
    auth_state = KalshiAuthState(
        api_key_id=require_kalshi_auth.api_key_id,
        private_key=require_kalshi_auth.private_key,
    )

    if _is_async_backend(backend_name):
        async with Client(base_url=KALSHI_DEMO_URL, inner=auth_state) as client:
            # Test that iterator works (may be empty if no fills)
            fills = await client.portfolio.get_fills_iter(limit=10).atake(5)
            # Just verify it returns a list (may be empty)
            assert isinstance(fills, list)
    else:
        with Client(base_url=KALSHI_DEMO_URL, inner=auth_state) as client:
            # Test that iterator works (may be empty if no fills)
            fills = client.portfolio.get_fills_iter(limit=10).take(5)
            # Just verify it returns a list (may be empty)
            assert isinstance(fills, list)



from .models import (
    GetItemsResponse,
    Item,
    ListOrdersResponse,
    ListUsersResponse,
    Order,
    User,
)

__all__ = [
    "GetItemsResponse",
    "Item",
    "ListOrdersResponse",
    "ListUsersResponse",
    "Order",
    "User",
]



"""Operation representations for OpenAPI operations."""

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any


class HttpMethod(Enum):
    """HTTP methods supported by OpenAPI."""

    GET = "get"
    POST = "post"
    PUT = "put"
    DELETE = "delete"
    PATCH = "patch"
    HEAD = "head"
    OPTIONS = "options"
    TRACE = "trace"


class ParameterLocation(Enum):
    """Where a parameter appears in the request."""

    PATH = "path"
    QUERY = "query"
    HEADER = "header"
    COOKIE = "cookie"


class ContentType(Enum):
    """Content types supported for request/response bodies."""

    JSON = "application/json"
    FORM = "application/x-www-form-urlencoded"
    MULTIPART = "multipart/form-data"
    BINARY = "application/octet-stream"
    TEXT = "text/plain"


@dataclass
class PathTemplate:
    """Represents a URL path template with parameters."""

    template: str
    parameters: list[str] = field(default_factory=lambda: [])

    @classmethod
    def from_openapi_path(cls, path: str) -> PathTemplate:
        """Parse an OpenAPI path template."""
        import re

        params = re.findall(r"\{([^}]+)\}", path)
        return cls(template=path, parameters=params)

    def to_python_fstring(self, param_mapping: dict[str, str] | None = None) -> str:
        """Convert to Python f-string format."""
        import re

        mapping = param_mapping or {}

        def replace(m: re.Match[str]) -> str:
            param = m.group(1)
            python_name = mapping.get(param, param)
            return f"{{_encode_path({python_name})}}"

        return re.sub(r"\{([^}]+)\}", replace, self.template)


@dataclass
class TypeId:
    """Unique identifier for a generated type."""

    name: str
    module: str = "types"

    def qualified_name(self) -> str:
        """Get the fully qualified type name."""
        return f"{self.module}.{self.name}" if self.module else self.name


@dataclass
class OperationParameter:
    """Parameter for an API operation."""

    name: str
    api_name: str
    location: ParameterLocation
    type_annotation: str
    required: bool
    description: str | None = None
    default: Any = None


@dataclass
class RequestBody:
    """Request body for an API operation."""

    content_type: ContentType
    type_annotation: str
    required: bool
    description: str | None = None


@dataclass
class OperationResponse:
    """Response for an API operation."""

    status_code: str
    content_type: ContentType | None
    type_annotation: str | None
    description: str | None = None


@dataclass
class OperationMethod:
    """Intermediate representation of an API operation."""

    operation_id: str
    tags: list[str]
    method: HttpMethod
    path: PathTemplate
    summary: str | None = None
    description: str | None = None
    parameters: list[OperationParameter] = field(default_factory=lambda: [])
    request_body: RequestBody | None = None
    responses: dict[str, OperationResponse] = field(default_factory=lambda: {})
    deprecated: bool = False
    pagination: PaginationInfo | None = None
    security: list[str] = field(default_factory=lambda: [])

    def python_method_name(self) -> str:
        """Get the Python method name for this operation."""
        import re

        name = self.operation_id
        # Convert camelCase/PascalCase to snake_case
        s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
        s2 = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1)
        # Replace non-alphanumeric with underscore
        name = re.sub(r"[^a-zA-Z0-9]", "_", s2)
        name = re.sub(r"_+", "_", name)
        name = name.strip("_").lower()
        return name if name else "operation"


@dataclass
class PaginationInfo:
    """Pagination metadata for operations that return paginated results."""

    item_type: str  # Type annotation for yielded items
    items_field: str  # Response field containing items array
    next_token_field: str  # Response field containing next token
    token_param: str  # Query param name for page token
    limit_param: str  # Query param name for limit
    token_type: str = "cursor"  # "cursor" or "url"


@dataclass
class WebSocketOperation:
    """Intermediate representation for WebSocket endpoints.

    WebSocket endpoints are marked with the x-websocket extension in OpenAPI.
    """

    operation_id: str
    path: PathTemplate
    parameters: list[OperationParameter] = field(default_factory=lambda: [])

    # Message types (from x-websocket-messages)
    send_type: str | None = None
    receive_type: str | None = None

    # Metadata
    tags: list[str] = field(default_factory=lambda: [])
    summary: str | None = None
    description: str | None = None
    deprecated: bool = False

    def python_method_name(self) -> str:
        """Get the Python method name for this operation."""
        import re

        name = self.operation_id
        # Convert camelCase/PascalCase to snake_case
        s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
        s2 = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1)
        # Replace non-alphanumeric with underscore
        name = re.sub(r"[^a-zA-Z0-9]", "_", s2)
        name = re.sub(r"_+", "_", name)
        name = name.strip("_").lower()
        return name if name else "ws_operation"



from __future__ import annotations
import msgspec
from enum import Enum
import datetime
import uuid
from typing import final, Any, Literal, TypeAlias
from collections.abc import Mapping, Sequence

DepthQuery: TypeAlias = int

FixedPointCount: TypeAlias = str

FixedPointDollars: TypeAlias = str

OrderbookLevel: TypeAlias = Sequence[float]

PriceLevel: TypeAlias = Sequence[int]

PriceLevelDollars: TypeAlias = Sequence[Any]

PriceLevelDollarsCountFp: TypeAlias = Sequence[str]


@final
class AcceptQuoteRequest(msgspec.Struct, frozen=True):
    accepted_side: Literal["yes", "no"]


@final
class AmendOrderRequest(msgspec.Struct, frozen=True):
    action: Literal["buy", "sell"]
    client_order_id: str
    side: Literal["yes", "no"]
    ticker: str
    updated_client_order_id: str
    count: int | None = None
    count_fp: FixedPointCount | None = None
    no_price: int | None = None
    no_price_dollars: FixedPointDollars | None = None
    yes_price: int | None = None
    yes_price_dollars: FixedPointDollars | None = None


@final
class AmendOrderResponse(msgspec.Struct, frozen=True):
    old_order: Order
    order: Order


@final
class Announcement(msgspec.Struct, frozen=True):
    delivery_time: datetime.datetime
    message: str
    status: Literal["active", "inactive"]
    type: Literal["info", "warning", "error"]


@final
class ApiKey(msgspec.Struct, frozen=True):
    api_key_id: str
    name: str
    scopes: Sequence[str]


@final
class ApplySubaccountTransferRequest(msgspec.Struct, frozen=True):
    amount_cents: int
    client_transfer_id: uuid.UUID
    from_subaccount: int
    to_subaccount: int


@final
class ApplySubaccountTransferResponse(msgspec.Struct, frozen=True):
    """Empty response indicating successful transfer."""


@final
class AssociatedEvent(msgspec.Struct, frozen=True):
    active_quoters: Sequence[str]
    is_yes_only: bool
    ticker: str
    size_max: int | None = None
    size_min: int | None = None


@final
class BatchCancelOrdersIndividualResponse(msgspec.Struct, frozen=True):
    order_id: str
    reduced_by: int
    reduced_by_fp: FixedPointCount
    error: ErrorResponse | None = None
    order: Order | None = None


@final
class BatchCancelOrdersRequest(msgspec.Struct, frozen=True):
    ids: Sequence[str]


@final
class BatchCancelOrdersResponse(msgspec.Struct, frozen=True):
    orders: Sequence[BatchCancelOrdersIndividualResponse]


@final
class BatchCreateOrdersIndividualResponse(msgspec.Struct, frozen=True):
    client_order_id: str | None = None
    error: ErrorResponse | None = None
    order: Order | None = None


@final
class BatchCreateOrdersRequest(msgspec.Struct, frozen=True):
    orders: Sequence[CreateOrderRequest]


@final
class BatchCreateOrdersResponse(msgspec.Struct, frozen=True):
    orders: Sequence[BatchCreateOrdersIndividualResponse]


@final
class BatchGetMarketCandlesticksResponse(msgspec.Struct, frozen=True):
    markets: Sequence[MarketCandlesticksResponse]


@final
class BidAskDistribution(msgspec.Struct, frozen=True):
    close: int
    close_dollars: FixedPointDollars
    high: int
    high_dollars: FixedPointDollars
    low: int
    low_dollars: FixedPointDollars
    open: int
    open_dollars: FixedPointDollars


@final
class CancelOrderResponse(msgspec.Struct, frozen=True):
    order: Order
    reduced_by: int
    reduced_by_fp: FixedPointCount


@final
class CreateApiKeyRequest(msgspec.Struct, frozen=True):
    name: str
    public_key: str
    scopes: Sequence[str] | None = None


@final
class CreateApiKeyResponse(msgspec.Struct, frozen=True):
    api_key_id: str


@final
class CreateMarketInMultivariateEventCollectionRequest(msgspec.Struct, frozen=True):
    selected_markets: Sequence[TickerPair]
    with_market_payload: bool | None = None


@final
class CreateMarketInMultivariateEventCollectionResponse(msgspec.Struct, frozen=True):
    event_ticker: str
    market_ticker: str
    market: Market | None = None


@final
class CreateOrderGroupRequest(msgspec.Struct, frozen=True):
    contracts_limit: int | None = None
    contracts_limit_fp: FixedPointCount | None = None


@final
class CreateOrderGroupResponse(msgspec.Struct, frozen=True):
    order_group_id: str


@final
class CreateOrderRequest(msgspec.Struct, frozen=True):
    action: Literal["buy", "sell"]
    side: Literal["yes", "no"]
    ticker: str
    buy_max_cost: int | None = None
    cancel_order_on_pause: bool | None = None
    client_order_id: str | None = None
    count: int | None = None
    count_fp: FixedPointCount | None = None
    expiration_ts: int | None = None
    no_price: int | None = None
    no_price_dollars: FixedPointDollars | None = None
    order_group_id: str | None = None
    post_only: bool | None = None
    reduce_only: bool | None = None
    self_trade_prevention_type: SelfTradePreventionType | None = None
    sell_position_floor: int | None = None
    subaccount: int | None = None
    time_in_force: (
        Literal["fill_or_kill", "good_till_canceled", "immediate_or_cancel"] | None
    ) = None
    type: Literal["limit", "market"] | None = None
    yes_price: int | None = None
    yes_price_dollars: FixedPointDollars | None = None


@final
class CreateOrderResponse(msgspec.Struct, frozen=True):
    order: Order


@final
class CreateQuoteRequest(msgspec.Struct, frozen=True):
    no_bid: FixedPointDollars
    rest_remainder: bool
    rfq_id: str
    yes_bid: FixedPointDollars


@final
class CreateQuoteResponse(msgspec.Struct, frozen=True):
    id: str


@final
class CreateRfqRequest(msgspec.Struct, frozen=True):
    market_ticker: str
    rest_remainder: bool
    contracts: int | None = None
    contracts_fp: FixedPointCount | None = None
    replace_existing: bool | None = None
    subtrader_id: str | None = None
    target_cost_centi_cents: int | None = None


@final
class CreateRfqResponse(msgspec.Struct, frozen=True):
    id: str


@final
class CreateSubaccountResponse(msgspec.Struct, frozen=True):
    subaccount_number: int


@final
class DailySchedule(msgspec.Struct, frozen=True):
    close_time: str
    open_time: str


@final
class DecreaseOrderRequest(msgspec.Struct, frozen=True):
    reduce_by: int | None = None
    reduce_by_fp: FixedPointCount | None = None
    reduce_to: int | None = None
    reduce_to_fp: FixedPointCount | None = None


@final
class DecreaseOrderResponse(msgspec.Struct, frozen=True):
    order: Order


@final
class EmptyResponse(msgspec.Struct, frozen=True):
    """An empty response body"""


@final
class ErrorResponse(msgspec.Struct, frozen=True):
    code: str | None = None
    details: str | None = None
    message: str | None = None
    service: str | None = None


@final
class EventData(msgspec.Struct, frozen=True):
    available_on_brokers: bool
    category: str
    collateral_return_type: str
    event_ticker: str
    mutually_exclusive: bool
    product_metadata: Mapping[str, Any]
    series_ticker: str
    sub_title: str
    title: str
    markets: Sequence[Market] | None = None
    strike_date: datetime.datetime | None = None
    strike_period: str | None = None


@final
class EventPosition(msgspec.Struct, frozen=True):
    event_exposure: int
    event_exposure_dollars: FixedPointDollars
    event_ticker: str
    fees_paid: int
    fees_paid_dollars: FixedPointDollars
    realized_pnl: int
    realized_pnl_dollars: FixedPointDollars
    total_cost: int
    total_cost_dollars: FixedPointDollars
    total_cost_shares: int
    total_cost_shares_fp: FixedPointCount


class ExchangeInstance(str, Enum):
    EVENT_CONTRACT = "event_contract"
    MARGINED = "margined"


@final
class ExchangeStatus(msgspec.Struct, frozen=True):
    exchange_active: bool
    trading_active: bool
    exchange_estimated_resume_time: datetime.datetime | None = None


@final
class Fill(msgspec.Struct, frozen=True):
    action: Literal["buy", "sell"]
    count: int
    count_fp: FixedPointCount
    fill_id: str
    is_taker: bool
    market_ticker: str
    no_price: int
    no_price_fixed: str
    order_id: str
    price: float
    side: Literal["yes", "no"]
    ticker: str
    trade_id: str
    yes_price: int
    yes_price_fixed: str
    client_order_id: str | None = None
    created_time: datetime.datetime | None = None
    ts: int | None = None


@final
class ForecastPercentilesPoint(msgspec.Struct, frozen=True):
    end_period_ts: int
    event_ticker: str
    percentile_points: Sequence[PercentilePoint]
    period_interval: int


@final
class GenerateApiKeyRequest(msgspec.Struct, frozen=True):
    name: str
    scopes: Sequence[str] | None = None


@final
class GenerateApiKeyResponse(msgspec.Struct, frozen=True):
    api_key_id: str
    private_key: str


@final
class GetApiKeysResponse(msgspec.Struct, frozen=True):
    api_keys: Sequence[ApiKey]


@final
class GetBalanceResponse(msgspec.Struct, frozen=True):
    balance: int
    portfolio_value: int
    updated_ts: int


@final
class GetCommunicationsIdResponse(msgspec.Struct, frozen=True):
    communications_id: str


@final
class GetEventCandlesticksResponse(msgspec.Struct, frozen=True):
    adjusted_end_ts: int
    market_candlesticks: Sequence[Sequence[MarketCandlestick]]
    market_tickers: Sequence[str]


@final
class GetEventForecastPercentilesHistoryResponse(msgspec.Struct, frozen=True):
    forecast_history: Sequence[ForecastPercentilesPoint]


@final
class GetEventMetadataResponse(msgspec.Struct, frozen=True):
    image_url: str
    market_details: Sequence[MarketMetadata]
    settlement_sources: Sequence[SettlementSource]
    competition: str | None = None
    competition_scope: str | None = None
    featured_image_url: str | None = None


@final
class GetEventResponse(msgspec.Struct, frozen=True):
    event: EventData
    markets: Sequence[Market]


@final
class GetEventsResponse(msgspec.Struct, frozen=True):
    cursor: str
    events: Sequence[EventData]
    milestones: Sequence[Milestone] | None = None


@final
class GetExchangeAnnouncementsResponse(msgspec.Struct, frozen=True):
    announcements: Sequence[Announcement]


@final
class GetExchangeScheduleResponse(msgspec.Struct, frozen=True):
    schedule: Schedule


@final
class GetFillsResponse(msgspec.Struct, frozen=True):
    cursor: str
    fills: Sequence[Fill]


@final
class GetFiltersBySportsResponse(msgspec.Struct, frozen=True):
    filters_by_sports: Mapping[str, SportFilterDetails]
    sport_ordering: Sequence[str]


@final
class GetIncentiveProgramsResponse(msgspec.Struct, frozen=True):
    incentive_programs: Sequence[IncentiveProgram]
    next_cursor: str | None = None


@final
class GetLiveDataResponse(msgspec.Struct, frozen=True):
    live_data: LiveData


@final
class GetLiveDatasResponse(msgspec.Struct, frozen=True):
    live_datas: Sequence[LiveData]


@final
class GetMarketCandlesticksResponse(msgspec.Struct, frozen=True):
    candlesticks: Sequence[MarketCandlestick]
    ticker: str


@final
class GetMarketOrderbookResponse(msgspec.Struct, frozen=True):
    orderbook: Orderbook
    orderbook_fp: OrderbookCountFp


@final
class GetMarketResponse(msgspec.Struct, frozen=True):
    market: Market


@final
class GetMarketsResponse(msgspec.Struct, frozen=True):
    cursor: str
    markets: Sequence[Market]


@final
class GetMilestoneResponse(msgspec.Struct, frozen=True):
    milestone: Milestone


@final
class GetMilestonesResponse(msgspec.Struct, frozen=True):
    milestones: Sequence[Milestone]
    cursor: str | None = None


@final
class GetMultivariateEventCollectionLookupHistoryResponse(msgspec.Struct, frozen=True):
    lookup_points: Sequence[LookupPoint]


@final
class GetMultivariateEventCollectionResponse(msgspec.Struct, frozen=True):
    multivariate_contract: MultivariateEventCollection


@final
class GetMultivariateEventCollectionsResponse(msgspec.Struct, frozen=True):
    multivariate_contracts: Sequence[MultivariateEventCollection]
    cursor: str | None = None


@final
class GetMultivariateEventsResponse(msgspec.Struct, frozen=True):
    cursor: str
    events: Sequence[EventData]


@final
class GetOrderGroupResponse(msgspec.Struct, frozen=True):
    is_auto_cancel_enabled: bool
    orders: Sequence[str]


@final
class GetOrderGroupsResponse(msgspec.Struct, frozen=True):
    order_groups: Sequence[OrderGroup] | None = None


@final
class GetOrderQueuePositionResponse(msgspec.Struct, frozen=True):
    queue_position: int


@final
class GetOrderQueuePositionsResponse(msgspec.Struct, frozen=True):
    queue_positions: Sequence[OrderQueuePosition]


@final
class GetOrderResponse(msgspec.Struct, frozen=True):
    order: Order


@final
class GetOrdersResponse(msgspec.Struct, frozen=True):
    cursor: str
    orders: Sequence[Order]


@final
class GetPortfolioRestingOrderTotalValueResponse(msgspec.Struct, frozen=True):
    total_resting_order_value: int


@final
class GetPositionsResponse(msgspec.Struct, frozen=True):
    event_positions: Sequence[EventPosition]
    market_positions: Sequence[MarketPosition]
    cursor: str | None = None


@final
class GetQuoteResponse(msgspec.Struct, frozen=True):
    quote: Quote


@final
class GetQuotesResponse(msgspec.Struct, frozen=True):
    quotes: Sequence[Quote]
    cursor: str | None = None


@final
class GetRfQsResponse(msgspec.Struct, frozen=True):
    rfqs: Sequence[Rfq]
    cursor: str | None = None


@final
class GetRfqResponse(msgspec.Struct, frozen=True):
    rfq: Rfq


@final
class GetSeriesFeeChangesResponse(msgspec.Struct, frozen=True):
    series_fee_change_arr: Sequence[SeriesFeeChange]


@final
class GetSeriesListResponse(msgspec.Struct, frozen=True):
    series: Sequence[Series]


@final
class GetSeriesResponse(msgspec.Struct, frozen=True):
    series: Series


@final
class GetSettlementsResponse(msgspec.Struct, frozen=True):
    settlements: Sequence[Settlement]
    cursor: str | None = None


@final
class GetStructuredTargetResponse(msgspec.Struct, frozen=True):
    structured_target: StructuredTarget | None = None


@final
class GetStructuredTargetsResponse(msgspec.Struct, frozen=True):
    cursor: str | None = None
    structured_targets: Sequence[StructuredTarget] | None = None


@final
class GetSubaccountBalancesResponse(msgspec.Struct, frozen=True):
    subaccount_balances: Sequence[SubaccountBalance]


@final
class GetSubaccountTransfersResponse(msgspec.Struct, frozen=True):
    transfers: Sequence[SubaccountTransfer]
    cursor: str | None = None


@final
class GetTagsForSeriesCategoriesResponse(msgspec.Struct, frozen=True):
    tags_by_categories: Mapping[str, Sequence[str]]


@final
class GetTradesResponse(msgspec.Struct, frozen=True):
    cursor: str
    trades: Sequence[Trade]


@final
class GetUserDataTimestampResponse(msgspec.Struct, frozen=True):
    as_of_time: datetime.datetime


@final
class IncentiveProgram(msgspec.Struct, frozen=True):
    end_date: datetime.datetime
    id: str
    incentive_type: Literal["liquidity", "volume"]
    market_ticker: str
    paid_out: bool
    period_reward: int
    start_date: datetime.datetime
    discount_factor_bps: int | None = None
    target_size: int | None = None
    target_size_fp: FixedPointCount | None = None


@final
class IntraExchangeInstanceTransferRequest(msgspec.Struct, frozen=True):
    amount: int
    destination: ExchangeInstance
    source: ExchangeInstance


@final
class IntraExchangeInstanceTransferResponse(msgspec.Struct, frozen=True):
    transfer_id: str


@final
class LiveData(msgspec.Struct, frozen=True):
    details: Mapping[str, Any]
    milestone_id: str
    type: str


@final
class LookupPoint(msgspec.Struct, frozen=True):
    event_ticker: str
    last_queried_ts: datetime.datetime
    market_ticker: str
    selected_markets: Sequence[TickerPair]


@final
class LookupTickersForMarketInMultivariateEventCollectionRequest(
    msgspec.Struct, frozen=True
):
    selected_markets: Sequence[TickerPair]


@final
class LookupTickersForMarketInMultivariateEventCollectionResponse(
    msgspec.Struct, frozen=True
):
    event_ticker: str
    market_ticker: str


@final
class MaintenanceWindow(msgspec.Struct, frozen=True):
    end_datetime: datetime.datetime
    start_datetime: datetime.datetime


@final
class Market(msgspec.Struct, frozen=True):
    can_close_early: bool
    close_time: datetime.datetime
    created_time: datetime.datetime
    event_ticker: str
    expiration_time: datetime.datetime
    expiration_value: str
    last_price: float
    last_price_dollars: FixedPointDollars
    latest_expiration_time: datetime.datetime
    liquidity: int
    liquidity_dollars: FixedPointDollars
    market_type: Literal["binary", "scalar"]
    no_ask: float
    no_ask_dollars: FixedPointDollars
    no_bid: float
    no_bid_dollars: FixedPointDollars
    no_sub_title: str
    notional_value: int
    notional_value_dollars: FixedPointDollars
    open_interest: int
    open_interest_fp: FixedPointCount
    open_time: datetime.datetime
    previous_price: int
    previous_price_dollars: FixedPointDollars
    previous_yes_ask: int
    previous_yes_ask_dollars: FixedPointDollars
    previous_yes_bid: int
    previous_yes_bid_dollars: FixedPointDollars
    price_level_structure: str
    price_ranges: Sequence[PriceRange]
    response_price_units: Literal["usd_cent"]
    result: Literal["yes", "no", ""]
    rules_primary: str
    rules_secondary: str
    settlement_timer_seconds: int
    status: Literal[
        "initialized",
        "inactive",
        "active",
        "closed",
        "determined",
        "disputed",
        "amended",
        "finalized",
    ]
    subtitle: str
    tick_size: int
    ticker: str
    title: str
    volume: int
    volume_24h: int
    volume_24h_fp: FixedPointCount
    volume_fp: FixedPointCount
    yes_ask: float
    yes_ask_dollars: FixedPointDollars
    yes_bid: float
    yes_bid_dollars: FixedPointDollars
    yes_sub_title: str
    cap_strike: float | None = None
    custom_strike: Mapping[str, Any] | None = None
    early_close_condition: str | None = None
    expected_expiration_time: datetime.datetime | None = None
    fee_waiver_expiration_time: datetime.datetime | None = None
    floor_strike: float | None = None
    functional_strike: str | None = None
    is_provisional: bool | None = None
    mve_collection_ticker: str | None = None
    mve_selected_legs: Sequence[MveSelectedLeg] | None = None
    primary_participant_key: str | None = None
    settlement_ts: datetime.datetime | None = None
    settlement_value: int | None = None
    settlement_value_dollars: FixedPointDollars | None = None
    strike_type: (
        Literal[
            "greater",
            "greater_or_equal",
            "less",
            "less_or_equal",
            "between",
            "functional",
            "custom",
            "structured",
        ]
        | None
    ) = None


@final
class MarketCandlestick(msgspec.Struct, frozen=True):
    end_period_ts: int
    open_interest: int
    open_interest_fp: FixedPointCount
    price: PriceDistribution
    volume: int
    volume_fp: FixedPointCount
    yes_ask: BidAskDistribution
    yes_bid: BidAskDistribution


@final
class MarketCandlesticksResponse(msgspec.Struct, frozen=True):
    candlesticks: Sequence[MarketCandlestick]
    market_ticker: str


@final
class MarketMetadata(msgspec.Struct, frozen=True):
    color_code: str
    image_url: str
    market_ticker: str


@final
class MarketPosition(msgspec.Struct, frozen=True):
    fees_paid: int
    fees_paid_dollars: FixedPointDollars
    market_exposure: int
    market_exposure_dollars: FixedPointDollars
    position: int
    position_fp: FixedPointCount
    realized_pnl: int
    realized_pnl_dollars: FixedPointDollars
    resting_orders_count: int
    ticker: str
    total_traded: int
    total_traded_dollars: FixedPointDollars
    last_updated_ts: datetime.datetime | None = None


@final
class Milestone(msgspec.Struct, frozen=True):
    category: str
    details: Mapping[str, Any]
    id: str
    last_updated_ts: datetime.datetime
    notification_message: str
    primary_event_tickers: Sequence[str]
    related_event_tickers: Sequence[str]
    start_date: datetime.datetime
    title: str
    type: str
    end_date: datetime.datetime | None = None
    source_id: str | None = None


@final
class MultivariateEventCollection(msgspec.Struct, frozen=True):
    associated_event_tickers: Sequence[str]
    associated_events: Sequence[AssociatedEvent]
    close_date: datetime.datetime
    collection_ticker: str
    description: str
    functional_description: str
    is_all_yes: bool
    is_ordered: bool
    is_single_market_per_event: bool
    open_date: datetime.datetime
    series_ticker: str
    size_max: int
    size_min: int
    title: str


@final
class MveSelectedLeg(msgspec.Struct, frozen=True):
    event_ticker: str | None = None
    market_ticker: str | None = None
    side: str | None = None
    yes_settlement_value_dollars: FixedPointDollars | None = None


@final
class Order(msgspec.Struct, frozen=True):
    action: Literal["buy", "sell"]
    client_order_id: str
    fill_count: int
    fill_count_fp: FixedPointCount
    initial_count: int
    initial_count_fp: FixedPointCount
    maker_fees: int
    maker_fill_cost: int
    maker_fill_cost_dollars: FixedPointDollars
    no_price: int
    no_price_dollars: FixedPointDollars
    order_id: str
    queue_position: int
    remaining_count: int
    remaining_count_fp: FixedPointCount
    side: Literal["yes", "no"]
    status: OrderStatus
    taker_fees: int
    taker_fill_cost: int
    taker_fill_cost_dollars: FixedPointDollars
    ticker: str
    type: Literal["limit", "market"]
    user_id: str
    yes_price: int
    yes_price_dollars: FixedPointDollars
    cancel_order_on_pause: bool | None = None
    created_time: datetime.datetime | None = None
    expiration_time: datetime.datetime | None = None
    last_update_time: datetime.datetime | None = None
    maker_fees_dollars: FixedPointDollars | None = None
    order_group_id: str | None = None
    self_trade_prevention_type: SelfTradePreventionType | None = None
    taker_fees_dollars: FixedPointDollars | None = None


@final
class OrderGroup(msgspec.Struct, frozen=True):
    id: str
    is_auto_cancel_enabled: bool


@final
class OrderQueuePosition(msgspec.Struct, frozen=True):
    market_ticker: str
    order_id: str
    queue_position: int


class OrderStatus(str, Enum):
    RESTING = "resting"
    CANCELED = "canceled"
    EXECUTED = "executed"


@final
class Orderbook(msgspec.Struct, frozen=True):
    """Legacy integer-count orderbook (will be deprecated). Prefer OrderbookCountFp / orderbook_fp for fixed-point contract counts."""

    no_dollars: Sequence[PriceLevelDollars]
    yes_dollars: Sequence[PriceLevelDollars]
    no: Sequence[OrderbookLevel] | None = None
    yes: Sequence[OrderbookLevel] | None = None


@final
class OrderbookCountFp(msgspec.Struct, frozen=True):
    """Orderbook with fixed-point contract counts (fp) in all dollar price levels."""

    no_dollars: Sequence[PriceLevelDollarsCountFp]
    yes_dollars: Sequence[PriceLevelDollarsCountFp]


@final
class PercentilePoint(msgspec.Struct, frozen=True):
    formatted_forecast: str
    numerical_forecast: float
    percentile: int
    raw_numerical_forecast: float


@final
class PriceDistribution(msgspec.Struct, frozen=True):
    close: int | None = None
    close_dollars: FixedPointDollars | None = None
    high: int | None = None
    high_dollars: FixedPointDollars | None = None
    low: int | None = None
    low_dollars: FixedPointDollars | None = None
    max: int | None = None
    max_dollars: FixedPointDollars | None = None
    mean: int | None = None
    mean_dollars: FixedPointDollars | None = None
    min: int | None = None
    min_dollars: FixedPointDollars | None = None
    open: int | None = None
    open_dollars: FixedPointDollars | None = None
    previous: int | None = None
    previous_dollars: FixedPointDollars | None = None


@final
class PriceRange(msgspec.Struct, frozen=True):
    end: str
    start: str
    step: str


@final
class Quote(msgspec.Struct, frozen=True):
    contracts: int
    contracts_fp: FixedPointCount
    created_ts: datetime.datetime
    creator_id: str
    id: str
    market_ticker: str
    no_bid: int
    no_bid_dollars: FixedPointDollars
    rfq_creator_id: str
    rfq_id: str
    status: Literal["open", "accepted", "confirmed", "executed", "cancelled"]
    updated_ts: datetime.datetime
    yes_bid: int
    yes_bid_dollars: FixedPointDollars
    accepted_side: Literal["yes", "no"] | None = None
    accepted_ts: datetime.datetime | None = None
    cancellation_reason: str | None = None
    cancelled_ts: datetime.datetime | None = None
    confirmed_ts: datetime.datetime | None = None
    creator_order_id: str | None = None
    creator_user_id: str | None = None
    executed_ts: datetime.datetime | None = None
    rest_remainder: bool | None = None
    rfq_creator_order_id: str | None = None
    rfq_creator_user_id: str | None = None
    rfq_target_cost_centi_cents: int | None = None


@final
class Rfq(msgspec.Struct, frozen=True):
    contracts: int
    contracts_fp: FixedPointCount
    created_ts: datetime.datetime
    creator_id: str
    id: str
    market_ticker: str
    status: Literal["open", "closed"]
    cancellation_reason: str | None = None
    cancelled_ts: datetime.datetime | None = None
    creator_user_id: str | None = None
    mve_collection_ticker: str | None = None
    mve_selected_legs: Sequence[MveSelectedLeg] | None = None
    rest_remainder: bool | None = None
    target_cost_centi_cents: int | None = None
    updated_ts: datetime.datetime | None = None


@final
class Schedule(msgspec.Struct, frozen=True):
    maintenance_windows: Sequence[MaintenanceWindow]
    standard_hours: Sequence[WeeklySchedule]


@final
class ScopeList(msgspec.Struct, frozen=True):
    scopes: Sequence[str]


class SelfTradePreventionType(str, Enum):
    TAKER_AT_CROSS = "taker_at_cross"
    MAKER = "maker"


@final
class Series(msgspec.Struct, frozen=True):
    additional_prohibitions: Sequence[str]
    category: str
    contract_terms_url: str
    contract_url: str
    fee_multiplier: float
    fee_type: Literal["quadratic", "quadratic_with_maker_fees", "flat"]
    frequency: str
    settlement_sources: Sequence[SettlementSource]
    tags: Sequence[str]
    ticker: str
    title: str
    product_metadata: Mapping[str, Any] | None = None
    volume: int | None = None
    volume_fp: FixedPointCount | None = None


@final
class SeriesFeeChange(msgspec.Struct, frozen=True):
    fee_multiplier: float
    fee_type: Literal["quadratic", "quadratic_with_maker_fees", "flat"]
    id: str
    scheduled_ts: datetime.datetime
    series_ticker: str


@final
class Settlement(msgspec.Struct, frozen=True):
    event_ticker: str
    fee_cost: str
    market_result: Literal["yes", "no", "scalar", "void"]
    no_count: int
    no_count_fp: FixedPointCount
    no_total_cost: int
    revenue: int
    settled_time: datetime.datetime
    ticker: str
    yes_count: int
    yes_count_fp: FixedPointCount
    yes_total_cost: int
    value: int | None = None


@final
class SettlementSource(msgspec.Struct, frozen=True):
    name: str | None = None
    url: str | None = None


@final
class SportFilterDetails(msgspec.Struct, frozen=True):
    competitions: Mapping[str, ScopeList]
    scopes: Sequence[str]


@final
class StructuredTarget(msgspec.Struct, frozen=True):
    details: Mapping[str, Any] | None = None
    id: str | None = None
    last_updated_ts: datetime.datetime | None = None
    name: str | None = None
    source_id: str | None = None
    type: str | None = None


@final
class SubaccountBalance(msgspec.Struct, frozen=True):
    balance: int
    subaccount_number: int
    updated_ts: int


@final
class SubaccountTransfer(msgspec.Struct, frozen=True):
    amount: int
    created_ts: int
    from_subaccount: int
    to_subaccount: int
    transfer_id: str


@final
class TickerPair(msgspec.Struct, frozen=True):
    event_ticker: str
    market_ticker: str
    side: Literal["yes", "no"]


@final
class Trade(msgspec.Struct, frozen=True):
    count: int
    count_fp: FixedPointCount
    no_price: int
    no_price_dollars: FixedPointDollars
    price: float
    taker_side: Literal["yes", "no"]
    ticker: str
    trade_id: str
    yes_price: int
    yes_price_dollars: FixedPointDollars
    created_time: datetime.datetime | None = None


@final
class WeeklySchedule(msgspec.Struct, frozen=True):
    end_time: datetime.datetime
    friday: Sequence[DailySchedule]
    monday: Sequence[DailySchedule]
    saturday: Sequence[DailySchedule]
    start_time: datetime.datetime
    sunday: Sequence[DailySchedule]
    thursday: Sequence[DailySchedule]
    tuesday: Sequence[DailySchedule]
    wednesday: Sequence[DailySchedule]



"""msgspec backend for dataclass generation."""

from __future__ import annotations

from typing import Any, Literal

import libcst as cst

from senzo.backends.base import (
    DataclassBackend,
    FieldDefinition,
    TypeMapping,
)
from senzo.settings import EnumStyle
from senzo.codegen.utils import make_annotation, make_docstring


class MsgspecBackend(DataclassBackend):
    """msgspec backend - fastest serialization, limited type support.

    Uses rename="camel" to automatically convert snake_case Python field names
    to camelCase for JSON serialization, matching typical API conventions.
    """

    def __init__(
        self,
        rename: Literal["camel", "pascal", "kebab", "screaming_snake", "lower", "upper"]
        | None = "camel",
        enum_style: EnumStyle | None = None,
    ) -> None:
        """Initialize msgspec backend.

        Args:
            rename: Renaming strategy for field names. Default is "camel" which
                   converts snake_case to camelCase. Set to None to disable.
            enum_style: How to generate enum types (LITERAL, ENUM, STR_ENUM, STR).
        """
        super().__init__(enum_style=enum_style)
        self._rename = rename

    def generate_class(
        self,
        name: str,
        fields: list[FieldDefinition],
        docstring: str | None = None,
    ) -> cst.ClassDef:
        """Generate a msgspec.Struct class with rename support."""
        body: list[cst.BaseStatement] = []

        if docstring:
            body.append(make_docstring(docstring))

        for field in fields:
            if field.required:
                ann_assign = cst.SimpleStatementLine(
                    body=[
                        cst.AnnAssign(
                            target=cst.Name(field.name),
                            annotation=make_annotation(field.type_annotation),
                            value=None,
                        )
                    ]
                )
            else:
                ann_assign = cst.SimpleStatementLine(
                    body=[
                        cst.AnnAssign(
                            target=cst.Name(field.name),
                            annotation=make_annotation(field.type_annotation),
                            value=cst.Name("None"),
                        )
                    ]
                )
            body.append(ann_assign)

        if not body:
            body.append(cst.SimpleStatementLine(body=[cst.Pass()]))

        # Build keyword arguments for Struct
        keywords: list[cst.Arg] = [
            cst.Arg(
                keyword=cst.Name("frozen"),
                value=cst.Name("True"),
            )
        ]
        if self._rename:
            keywords.append(
                cst.Arg(
                    keyword=cst.Name("rename"),
                    value=cst.SimpleString(f'"{self._rename}"'),
                )
            )

        return cst.ClassDef(
            name=cst.Name(name),
            bases=[
                cst.Arg(
                    value=cst.Attribute(
                        value=cst.Name("msgspec"),
                        attr=cst.Name("Struct"),
                    )
                ),
            ],
            keywords=keywords,
            body=cst.IndentedBlock(body=body),
            decorators=[cst.Decorator(decorator=cst.Name("final"))],
        )

    def generate_enum(
        self,
        name: str,
        values: list[tuple[str, Any]],
        style: EnumStyle | None = None,
    ) -> cst.ClassDef:
        """Generate an enum class.

        Args:
            name: The enum class name
            values: List of (name, value) tuples for enum members
            style: Override enum style, or use backend default
        """
        effective_style = style or self._enum_style
        body: list[cst.BaseStatement] = []

        for enum_name, enum_value in values:
            safe_name = self._make_enum_name(str(enum_name))
            if isinstance(enum_value, str):
                value_expr: cst.BaseExpression = cst.SimpleString(f'"{enum_value}"')
            else:
                value_expr = cst.parse_expression(repr(enum_value))

            body.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.Assign(
                            targets=[cst.AssignTarget(target=cst.Name(safe_name))],
                            value=value_expr,
                        )
                    ]
                )
            )

        if not body:
            body.append(cst.SimpleStatementLine(body=[cst.Pass()]))

        # Choose base class(es) based on enum style
        if effective_style == EnumStyle.STR_ENUM:
            bases = [cst.Arg(value=cst.Name("StrEnum"))]
        elif effective_style == EnumStyle.ENUM:
            bases = [cst.Arg(value=cst.Name("str")), cst.Arg(value=cst.Name("Enum"))]
        else:
            bases = [cst.Arg(value=cst.Name("str")), cst.Arg(value=cst.Name("Enum"))]

        return cst.ClassDef(
            name=cst.Name(name),
            bases=bases,
            body=cst.IndentedBlock(body=body),
        )

    def _make_enum_name(self, value: str) -> str:
        """Convert a value to a valid Python enum name."""
        import re

        name = re.sub(r"[^a-zA-Z0-9_]", "_", value)
        if name and name[0].isdigit():
            name = f"_{name}"
        return name.upper() or "UNKNOWN"

    def get_imports(self) -> list[cst.SimpleStatementLine]:
        """Return required imports for msgspec."""
        imports: list[cst.SimpleStatementLine] = [
            cst.SimpleStatementLine(
                body=[cst.Import(names=[cst.ImportAlias(name=cst.Name("msgspec"))])]
            ),
        ]

        # Enum imports based on style
        if self._enum_style == EnumStyle.STR_ENUM:
            imports.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.ImportFrom(
                            module=cst.Name("enum"),
                            names=[cst.ImportAlias(name=cst.Name("StrEnum"))],
                        )
                    ]
                )
            )
        else:
            imports.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.ImportFrom(
                            module=cst.Name("enum"),
                            names=[cst.ImportAlias(name=cst.Name("Enum"))],
                        )
                    ]
                )
            )

        return imports

    def get_type_mapping(self) -> TypeMapping:
        """Return type mappings for OpenAPI types."""
        return TypeMapping(
            string="str",
            integer="int",
            number="float",
            boolean="bool",
            date="datetime.date",
            datetime="datetime.datetime",
            uuid="uuid.UUID",
            email="str",
            uri="str",
            binary="bytes",
        )

    def supports_type(self, openapi_type: str, format_: str | None) -> bool:
        """Check if backend supports a specific type."""
        supported_types = {"string", "integer", "number", "boolean", "array", "object"}
        supported_formats = {"date", "date-time", "uuid", "email", "uri", "binary"}
        if openapi_type not in supported_types:
            return False
        if format_ and format_ not in supported_formats:
            return False
        return True

    def generate_decode_json_expression(
        self,
        type_name: str,
        bytes_expr: cst.BaseExpression,
    ) -> cst.BaseExpression:
        """Generate msgspec.json.decode expression.

        Args:
            type_name: The type to decode to (e.g., "PostResponse" or "list[Pet]")
            bytes_expr: Expression that evaluates to bytes (e.g., _response.content)

        Returns:
            msgspec.json.decode(bytes_expr, type=TypeName)
        """
        return cst.Call(
            func=cst.Attribute(
                value=cst.Attribute(
                    value=cst.Name("msgspec"),
                    attr=cst.Name("json"),
                ),
                attr=cst.Name("decode"),
            ),
            args=[
                cst.Arg(value=bytes_expr),
                cst.Arg(
                    keyword=cst.Name("type"),
                    value=cst.parse_expression(type_name),
                ),
            ],
        )

    def generate_encode_json_expression(
        self,
        type_annotation: str,
        body_expr: cst.BaseExpression,
    ) -> cst.BaseExpression:
        """Generate msgspec.json.encode expression.

        Args:
            type_annotation: The type annotation (unused for msgspec)
            body_expr: Expression for the object to encode

        Returns:
            msgspec.json.encode(body_expr)
        """
        del type_annotation  # msgspec.json.encode handles any type
        return cst.Call(
            func=cst.Attribute(
                value=cst.Attribute(
                    value=cst.Name("msgspec"),
                    attr=cst.Name("json"),
                ),
                attr=cst.Name("encode"),
            ),
            args=[cst.Arg(value=body_expr)],
        )

    def get_client_imports(self) -> list[cst.SimpleStatementLine]:
        """Return imports needed in client.py for msgspec encoding/decoding."""
        return [
            cst.SimpleStatementLine(
                body=[cst.Import(names=[cst.ImportAlias(name=cst.Name("msgspec"))])]
            ),
            cst.SimpleStatementLine(
                body=[
                    cst.ImportFrom(
                        module=cst.Name("typing"),
                        names=[cst.ImportAlias(name=cst.Name("Any"))],
                    )
                ]
            ),
        ]



"""TypeSpace: Schema to type conversion and management."""

from __future__ import annotations

import logging
import re
from dataclasses import dataclass, field
from typing import Any, cast

import libcst as cst

from collections.abc import Callable

from senzo.backends.base import DataclassBackend, FieldDefinition
from senzo.settings import EnumStyle
from senzo.codegen.utils import make_module
from senzo.operation import TypeId

logger = logging.getLogger(__name__)

# Type alias for a function that resolves $ref to a schema dict
RefResolver = Callable[[str], dict[str, Any]]

# Python reserved keywords - suffix with _ per PEP 8
PYTHON_KEYWORDS = frozenset(
    {
        "False",
        "None",
        "True",
        "and",
        "as",
        "assert",
        "async",
        "await",
        "break",
        "class",
        "continue",
        "def",
        "del",
        "elif",
        "else",
        "except",
        "finally",
        "for",
        "from",
        "global",
        "if",
        "import",
        "in",
        "is",
        "lambda",
        "nonlocal",
        "not",
        "or",
        "pass",
        "raise",
        "return",
        "try",
        "while",
        "with",
        "yield",
    }
)


def to_snake_case(name: str | bool) -> str:
    """Convert a string to snake_case.

    Handles YAML boolean coercion (yes/no -> True/False).
    Suffixes Python keywords with _ per PEP 8.
    """
    # Handle YAML boolean coercion: yes/no -> True/False
    if isinstance(name, bool):
        name = "yes" if name else "no"
    s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    s2 = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1)
    result = re.sub(r"[^a-zA-Z0-9]", "_", s2).lower()
    if result in PYTHON_KEYWORDS or result.capitalize() in PYTHON_KEYWORDS:
        result = f"{result}_"
    return result


def to_pascal_case(name: str) -> str:
    """Convert a string to PascalCase."""
    s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    s2 = re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1)
    words = re.split(r"[^a-zA-Z0-9]", s2)
    return "".join(word.capitalize() for word in words if word)


@dataclass
class NormalizedSchema:
    """Normalized schema after resolving refs and composition.

    This is an intermediate representation used to handle allOf, oneOf, anyOf
    before generating types.
    """

    properties: dict[str, dict[str, Any]] = field(
        default_factory=lambda: cast(dict[str, dict[str, Any]], {})
    )
    required: set[str] = field(default_factory=lambda: cast(set[str], set()))
    description: str | None = None
    additional_properties: bool | dict[str, Any] | None = None
    is_enum: bool = False
    enum_values: list[Any] = field(default_factory=lambda: cast(list[Any], []))


@dataclass
class TypeDefinition:
    """A generated type definition."""

    type_id: TypeId
    schema: dict[str, Any]
    normalized: NormalizedSchema | None = None
    class_def: cst.ClassDef | None = None
    is_anonymous: bool = False
    is_alias: bool = False
    alias_target: str | None = None


class TypeSpace:
    """Manages type generation and deduplication."""

    def __init__(
        self,
        backend: DataclassBackend,
        ref_resolver: RefResolver | None = None,
        inline_enum_style: EnumStyle | None = None,
    ) -> None:
        self._backend = backend
        self._types: dict[str, TypeDefinition] = {}
        self._refs: dict[str, TypeId] = {}
        self._type_mapping = backend.get_type_mapping()
        self._pending_refs: set[str] = set()
        self._uses_any: bool = False
        self._uses_literal: bool = False
        self._uses_datetime: bool = False
        self._uses_uuid: bool = False
        self._uses_type_alias: bool = False
        self._uses_mapping: bool = False
        self._uses_sequence: bool = False
        self._for_input: bool = False
        self._ref_resolver = ref_resolver
        # Default inline enums to Literal if not specified
        self._inline_enum_style = inline_enum_style or EnumStyle.LITERAL
        # Counter for generating anonymous type names
        self._anon_counter: int = 0
        # Context hint for anonymous type naming (set by caller)
        self._context_hint: str | None = None

    def add_schema(
        self,
        name: str,
        schema: dict[str, Any],
        is_anonymous: bool = False,
    ) -> TypeId:
        """Convert and register a schema, returning its TypeId."""
        pascal_name = to_pascal_case(name)
        type_id = TypeId(name=pascal_name)

        if pascal_name in self._types:
            return self._types[pascal_name].type_id

        self._types[pascal_name] = TypeDefinition(
            type_id=type_id,
            schema=schema,
            is_anonymous=is_anonymous,
        )

        return type_id

    def register_inline_response_schema(
        self,
        operation_id: str,
        schema: dict[str, Any],
    ) -> str:
        """Register an inline response schema and return its type name.

        Generates a name like `{OperationId}Response` for inline object schemas.
        Reuses existing types if an equivalent schema is already registered.
        """
        return self._register_inline_schema(operation_id, schema, suffix="Response")

    def register_inline_request_schema(
        self,
        operation_id: str,
        schema: dict[str, Any],
        suffix: str = "Request",
    ) -> str:
        """Register an inline request schema and return its type name.

        Generates a name like `{OperationId}Request` for inline object schemas.
        Reuses existing types if an equivalent schema is already registered.
        """
        return self._register_inline_schema(operation_id, schema, suffix=suffix)

    def _register_inline_schema(
        self,
        operation_id: str,
        schema: dict[str, Any],
        suffix: str,
    ) -> str:
        """Register an inline schema and return its type name."""
        # Check if an equivalent schema already exists
        schema_hash = self._hash_schema(schema)
        for type_name, type_def in self._types.items():
            if self._hash_schema(type_def.schema) == schema_hash:
                return type_name

        type_name = f"{to_pascal_case(operation_id)}{suffix}"

        # Handle name conflicts by appending a counter
        base_name = type_name
        counter = 2
        while type_name in self._types:
            type_name = f"{base_name}{counter}"
            counter += 1

        self.add_schema(type_name, schema, is_anonymous=True)
        return type_name

    def _hash_schema(self, schema: dict[str, Any]) -> str:
        """Create a hash of a schema for deduplication.

        Ignores description/example fields, focuses on structural properties.
        """
        import hashlib
        import json

        def normalize(obj: object) -> object:
            if isinstance(obj, dict):
                # Skip non-structural fields
                skip = {"description", "example", "title"}
                d = cast(dict[str, object], obj)
                return {
                    k: normalize(v)
                    for k, v in sorted(d.items())
                    if k not in skip and not k.startswith("x-")
                }
            if isinstance(obj, list):
                return [normalize(item) for item in cast(list[object], obj)]
            return obj

        normalized = normalize(schema)
        return hashlib.md5(json.dumps(normalized, sort_keys=True).encode()).hexdigest()

    def resolve_ref(self, ref: str) -> TypeId:
        """Resolve a $ref to its TypeId."""
        if ref in self._refs:
            return self._refs[ref]

        parts = ref.split("/")
        name = parts[-1]
        type_id = TypeId(name=to_pascal_case(name))
        self._refs[ref] = type_id
        return type_id

    def get_type_annotation(
        self, schema: dict[str, Any], *, for_input: bool = False
    ) -> str:
        """Get Python type annotation for a schema.

        Args:
            schema: The OpenAPI schema to convert.
            for_input: If True, use mutable types (dict) for inputs.
                      If False, use immutable types (Mapping) for outputs.
        """
        self._for_input = for_input
        if "$ref" in schema:
            type_id = self.resolve_ref(schema["$ref"])
            return type_id.name

        openapi_type = schema.get("type")
        format_ = schema.get("format")

        # Handle inline enums (with or without explicit type)
        if "enum" in schema:
            return self._handle_inline_enum(schema)

        if openapi_type is None:
            if "allOf" in schema:
                return self._handle_all_of(schema)
            if "oneOf" in schema:
                return self._handle_one_of(schema)
            if "anyOf" in schema:
                return self._handle_any_of(schema)
            self._uses_any = True
            return "Any"

        format_type = self._type_mapping.get_format_type(openapi_type, format_)
        if format_type:
            if format_type.startswith("datetime."):
                self._uses_datetime = True
            elif format_type.startswith("uuid."):
                self._uses_uuid = True
            return format_type

        return self._get_base_type(openapi_type, schema)

    def _handle_inline_enum(self, schema: dict[str, Any]) -> str:
        """Handle inline enum schemas based on configured enum style.

        Args:
            schema: Schema with 'enum' key

        Returns:
            Type annotation string (Literal, str, or reference to named enum)
        """
        enum_values = schema.get("enum", [])
        if not enum_values:
            return "str"

        if self._inline_enum_style == EnumStyle.LITERAL:
            self._uses_literal = True
            return self._backend.generate_literal_type(enum_values)
        elif self._inline_enum_style == EnumStyle.STR:
            return "str"
        else:
            # For ENUM or STR_ENUM styles on inline enums, fall back to str
            # since we can't generate a named enum class inline
            return "str"

    def _get_base_type(self, openapi_type: str, schema: dict[str, Any]) -> str:
        """Get base Python type for an OpenAPI type."""
        type_map: dict[str, str] = {
            "string": self._type_mapping.string,
            "integer": self._type_mapping.integer,
            "number": self._type_mapping.number,
            "boolean": self._type_mapping.boolean,
            "null": "None",
        }

        if openapi_type in type_map:
            return type_map[openapi_type]

        if openapi_type == "array":
            items = schema.get("items", {})
            # Use list for inputs (mutable), Sequence for outputs (immutable)
            is_input = self._for_input
            item_type = self.get_type_annotation(items, for_input=is_input)
            if is_input:
                return f"list[{item_type}]"
            else:
                self._uses_sequence = True
                return f"Sequence[{item_type}]"

        if openapi_type == "object":
            additional_props = schema.get("additionalProperties")
            # Use dict for inputs (mutable), Mapping for outputs (immutable)
            is_input = self._for_input
            map_type = "dict" if is_input else "Mapping"
            if not is_input:
                self._uses_mapping = True
            if additional_props is True:
                self._uses_any = True
                return f"{map_type}[str, Any]"
            if isinstance(additional_props, dict):
                value_type = self.get_type_annotation(
                    cast(dict[str, Any], additional_props), for_input=is_input
                )
                return f"{map_type}[str, {value_type}]"
            if "properties" in schema:
                self._uses_any = True
                return f"{map_type}[str, Any]"
            self._uses_any = True
            return f"{map_type}[str, Any]"

        self._uses_any = True
        return "Any"

    def _handle_all_of(self, schema: dict[str, Any]) -> str:
        """Handle allOf composition by merging schemas.

        For inline allOf with properties, we generate an anonymous type and
        register it so it gets generated with other types.
        """
        all_of: list[dict[str, Any]] = schema.get("allOf", [])
        if not all_of:
            self._uses_any = True
            return "Any"

        if len(all_of) == 1:
            return self.get_type_annotation(all_of[0])

        # Check what kind of schemas we have
        ref_types: list[str] = []
        has_inline_properties = False

        for s in all_of:
            if "$ref" in s:
                ref_types.append(self.get_type_annotation(s))
            elif s.get("properties") or s.get("type") == "object":
                has_inline_properties = True

        # If it's all refs and no inline properties, return the first one
        if ref_types and not has_inline_properties:
            return ref_types[0]

        # If there are inline properties, generate an anonymous merged type
        if has_inline_properties:
            anon_name = self._generate_anonymous_type_name()
            self._register_anonymous_type(anon_name, schema)
            return anon_name

        # Fall back to first type
        types = [self.get_type_annotation(s) for s in all_of]
        return types[0] if types else "Any"

    def _generate_anonymous_type_name(self) -> str:
        """Generate a unique name for an anonymous type.

        Uses context hint if available, otherwise falls back to counter.
        """
        self._anon_counter += 1
        if self._context_hint:
            name = to_pascal_case(self._context_hint)
            # Clear context after use
            self._context_hint = None
            return name
        return f"_AnonType{self._anon_counter}"

    def _register_anonymous_type(self, name: str, schema: dict[str, Any]) -> None:
        """Register an anonymous type so it gets generated with other types."""
        if name in self._types:
            return

        type_id = TypeId(name=name)
        self._types[name] = TypeDefinition(
            type_id=type_id,
            schema=schema,
            is_anonymous=True,
        )

    def set_context_hint(self, hint: str) -> None:
        """Set a context hint for naming the next anonymous type.

        This allows callers to provide meaningful names like 'CreateUserRequest'
        instead of '_AnonType1'.
        """
        self._context_hint = hint

    def normalize_schema(self, schema: dict[str, Any]) -> NormalizedSchema:
        """Normalize a schema by resolving $refs and merging allOf.

        This creates an intermediate representation suitable for type generation.
        """
        normalized = NormalizedSchema()
        normalized.description = schema.get("description")

        # Handle enum schemas
        if "enum" in schema:
            normalized.is_enum = True
            normalized.enum_values = schema["enum"]
            return normalized

        # Handle allOf composition
        if "allOf" in schema:
            return self._normalize_all_of(schema)

        # Handle $ref
        if "$ref" in schema:
            resolved = self._resolve_ref_schema(schema["$ref"])
            if resolved:
                return self.normalize_schema(resolved)
            return normalized

        # Handle regular object schema
        normalized.properties = dict(schema.get("properties", {}))
        normalized.required = set(schema.get("required", []))
        normalized.additional_properties = schema.get("additionalProperties")

        return normalized

    def _normalize_all_of(self, schema: dict[str, Any]) -> NormalizedSchema:
        """Normalize an allOf schema by merging all subschemas."""
        all_of: list[dict[str, Any]] = schema.get("allOf", [])
        merged = NormalizedSchema()
        merged.description = schema.get("description")

        for subschema in all_of:
            sub_normalized = self.normalize_schema(subschema)

            # Merge properties (last wins on conflict, with warning)
            for prop_name, prop_schema in sub_normalized.properties.items():
                if prop_name in merged.properties:
                    logger.warning(
                        "Property '%s' defined in multiple allOf subschemas, "
                        "using last definition",
                        prop_name,
                    )
                merged.properties[prop_name] = prop_schema

            # Merge required fields
            merged.required.update(sub_normalized.required)

            # Merge description (prefer first non-None)
            if merged.description is None and sub_normalized.description:
                merged.description = sub_normalized.description

            # Handle additional properties (last non-None wins)
            if sub_normalized.additional_properties is not None:
                merged.additional_properties = sub_normalized.additional_properties

        return merged

    def _resolve_ref_schema(self, ref: str) -> dict[str, Any] | None:
        """Resolve a $ref to its schema dict.

        Returns None if no resolver is configured or ref not found.
        """
        if self._ref_resolver is None:
            # Try to look it up in registered types
            parts = ref.split("/")
            name = to_pascal_case(parts[-1])
            if name in self._types:
                return self._types[name].schema
            return None

        return self._ref_resolver(ref)

    def _handle_one_of(self, schema: dict[str, Any]) -> str:
        """Handle oneOf composition."""
        one_of = schema.get("oneOf", [])
        types = [self.get_type_annotation(s) for s in one_of]
        return " | ".join(types) if types else "Any"

    def _handle_any_of(self, schema: dict[str, Any]) -> str:
        """Handle anyOf composition."""
        any_of = schema.get("anyOf", [])
        types = [self.get_type_annotation(s) for s in any_of]
        return " | ".join(types) if types else "Any"

    def _schema_to_fields(self, schema: dict[str, Any]) -> list[FieldDefinition]:
        """Convert a schema's properties to field definitions."""
        properties = schema.get("properties", {})
        required_fields = set(schema.get("required", []))
        fields: list[FieldDefinition] = []

        for prop_name, prop_schema in properties.items():
            # Handle YAML boolean coercion: yes/no -> True/False
            api_name = (
                "yes"
                if prop_name is True
                else "no"
                if prop_name is False
                else prop_name
            )
            python_name = to_snake_case(prop_name)
            type_annotation = self.get_type_annotation(prop_schema)
            is_required = prop_name in required_fields

            if not is_required:
                type_annotation = f"{type_annotation} | None"

            fields.append(
                FieldDefinition(
                    name=python_name,
                    api_name=api_name,
                    type_annotation=type_annotation,
                    required=is_required,
                    default=None if is_required else "None",
                    description=prop_schema.get("description"),
                )
            )

        fields.sort(key=lambda f: (not f.required, f.name))
        return fields

    def _normalized_to_fields(
        self, normalized: NormalizedSchema
    ) -> list[FieldDefinition]:
        """Convert a normalized schema's properties to field definitions."""
        fields: list[FieldDefinition] = []

        for prop_name, prop_schema in normalized.properties.items():
            # NormalizedSchema properties already have string keys
            python_name = to_snake_case(prop_name)
            type_annotation = self.get_type_annotation(prop_schema)
            is_required = prop_name in normalized.required

            if not is_required:
                type_annotation = f"{type_annotation} | None"

            fields.append(
                FieldDefinition(
                    name=python_name,
                    api_name=prop_name,
                    type_annotation=type_annotation,
                    required=is_required,
                    default=None if is_required else "None",
                    description=prop_schema.get("description"),
                )
            )

        fields.sort(key=lambda f: (not f.required, f.name))
        return fields

    def generate_type(self, name: str) -> cst.ClassDef | cst.SimpleStatementLine | None:
        """Generate a class definition or type alias for a registered type."""
        if name not in self._types:
            return None

        type_def = self._types[name]
        schema = type_def.schema

        # Handle enum schemas
        if "enum" in schema:
            values = [(str(v), v) for v in schema["enum"]]
            class_def = self._backend.generate_enum(name, values)
            type_def.class_def = class_def
            return class_def

        # Handle allOf schemas by normalizing first
        if "allOf" in schema:
            normalized = self.normalize_schema(schema)
            type_def.normalized = normalized

            if normalized.is_enum:
                values = [(str(v), v) for v in normalized.enum_values]
                class_def = self._backend.generate_enum(name, values)
            else:
                fields = self._normalized_to_fields(normalized)
                docstring = normalized.description
                class_def = self._backend.generate_class(name, fields, docstring)

            type_def.class_def = class_def
            return class_def

        # Handle primitive type aliases (string, integer, etc. without properties)
        openapi_type = schema.get("type")
        has_properties = bool(schema.get("properties"))

        if openapi_type and not has_properties and openapi_type != "object":
            # This is a primitive type alias (e.g., FixedPointDollars = str)
            base_type = self.get_type_annotation(schema)
            type_def.is_alias = True
            type_def.alias_target = base_type
            return self._generate_type_alias(name, base_type, schema.get("description"))

        # Handle regular schemas with properties
        fields = self._schema_to_fields(schema)
        docstring = schema.get("description")
        class_def = self._backend.generate_class(name, fields, docstring)
        type_def.class_def = class_def
        return class_def

    def _generate_type_alias(
        self, name: str, target: str, description: str | None = None
    ) -> cst.SimpleStatementLine:
        """Generate a type alias statement.

        Generates: TypeName: TypeAlias = TargetType
        """
        self._uses_type_alias = True
        # Parse the target type string into a CST expression
        target_expr = cst.parse_expression(target)
        return cst.SimpleStatementLine(
            body=[
                cst.AnnAssign(
                    target=cst.Name(name),
                    annotation=cst.Annotation(annotation=cst.Name("TypeAlias")),
                    value=target_expr,
                )
            ],
            leading_lines=[cst.EmptyLine()] if description else [],
        )

    def generate_types_module(self) -> cst.Module:
        """Generate the types module with all registered types."""
        statements: list[cst.SimpleStatementLine | cst.BaseCompoundStatement] = []

        # Add __future__ annotations first (must be first import)
        statements.append(
            cst.SimpleStatementLine(
                body=[
                    cst.ImportFrom(
                        module=cst.Name("__future__"),
                        names=[cst.ImportAlias(name=cst.Name("annotations"))],
                    )
                ]
            )
        )

        statements.extend(self._backend.get_imports())

        # Generate all types first to populate usage flags
        type_aliases: list[cst.SimpleStatementLine] = []
        class_defs: list[cst.ClassDef] = []
        for name in sorted(self._types.keys()):
            type_def = self.generate_type(name)
            if isinstance(type_def, cst.SimpleStatementLine):
                type_aliases.append(type_def)
            elif isinstance(type_def, cst.ClassDef):
                class_defs.append(type_def)

        # Add standard library imports if needed
        if self._uses_datetime:
            statements.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.Import(names=[cst.ImportAlias(name=cst.Name("datetime"))])
                    ]
                )
            )
        if self._uses_uuid:
            statements.append(
                cst.SimpleStatementLine(
                    body=[cst.Import(names=[cst.ImportAlias(name=cst.Name("uuid"))])]
                )
            )

        # Add typing imports if needed
        typing_imports: list[cst.ImportAlias] = [
            # Always import final for @final decorator on classes
            cst.ImportAlias(name=cst.Name("final")),
        ]
        if self._uses_any:
            typing_imports.append(cst.ImportAlias(name=cst.Name("Any")))
        if self._uses_literal:
            typing_imports.append(cst.ImportAlias(name=cst.Name("Literal")))
        if self._uses_type_alias:
            typing_imports.append(cst.ImportAlias(name=cst.Name("TypeAlias")))

        if typing_imports:
            statements.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.ImportFrom(
                            module=cst.Name("typing"),
                            names=typing_imports,
                        )
                    ]
                )
            )

        # Add collections.abc imports if needed
        collections_abc_imports: list[cst.ImportAlias] = []
        if self._uses_mapping:
            collections_abc_imports.append(cst.ImportAlias(name=cst.Name("Mapping")))
        if self._uses_sequence:
            collections_abc_imports.append(cst.ImportAlias(name=cst.Name("Sequence")))
        if collections_abc_imports:
            statements.append(
                cst.SimpleStatementLine(
                    body=[
                        cst.ImportFrom(
                            module=cst.Attribute(
                                value=cst.Name("collections"),
                                attr=cst.Name("abc"),
                            ),
                            names=collections_abc_imports,
                        )
                    ]
                )
            )

        # Add type aliases before classes (they may be referenced by classes)
        statements.extend(type_aliases)
        statements.extend(class_defs)

        return make_module(statements)

    def get_registered_types(self) -> list[str]:
        """Get list of registered type names."""
        return sorted(self._types.keys())

    def get_schema_for_type(self, type_name: str) -> dict[str, Any] | None:
        """Get the original schema for a registered type name."""
        if type_name in self._types:
            return self._types[type_name].schema
        return None

    @property
    def uses_literal(self) -> bool:
        """Check if any type annotation uses Literal."""
        return self._uses_literal

    @property
    def uses_datetime(self) -> bool:
        """Check if any type annotation uses datetime."""
        return self._uses_datetime

    @property
    def uses_uuid(self) -> bool:
        """Check if any type annotation uses uuid."""
        return self._uses_uuid

    @property
    def uses_mapping(self) -> bool:
        """Check if any type annotation uses Mapping."""
        return self._uses_mapping

    @property
    def uses_sequence(self) -> bool:
        """Check if any type annotation uses Sequence."""
        return self._uses_sequence



"""OpenAPI specification parser and validator."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any, cast

import yaml
from openapi_spec_validator import validate
from openapi_spec_validator.readers import read_from_filename


class OpenAPISpec:
    """Parsed and validated OpenAPI specification."""

    def __init__(self, spec: dict[str, Any]) -> None:
        self._spec = spec

    @classmethod
    def from_dict(cls, spec: dict[str, Any]) -> OpenAPISpec:
        """Create from a dictionary, validating the spec."""
        validate(spec)  # type: ignore[arg-type]
        return cls(spec)

    @classmethod
    def from_file(cls, path: str | Path) -> OpenAPISpec:
        """Load and validate from a file path."""
        spec_dict, _ = read_from_filename(str(path))
        spec_dict = cast(dict[str, Any], spec_dict)
        validate(spec_dict)  # type: ignore[arg-type]
        return cls(spec_dict)

    @classmethod
    def from_json(cls, content: str) -> OpenAPISpec:
        """Parse from JSON string."""
        spec = json.loads(content)
        return cls.from_dict(spec)

    @classmethod
    def from_yaml(cls, content: str) -> OpenAPISpec:
        """Parse from YAML string."""
        spec = yaml.safe_load(content)
        return cls.from_dict(spec)

    @property
    def raw(self) -> dict[str, Any]:
        """Get the raw spec dictionary."""
        return self._spec

    @property
    def openapi_version(self) -> str:
        """Get the OpenAPI version."""
        return str(self._spec.get("openapi", "3.0.0"))

    @property
    def info(self) -> dict[str, Any]:
        """Get the info object."""
        return dict(self._spec.get("info", {}))

    @property
    def title(self) -> str:
        """Get the API title."""
        return str(self.info.get("title", "API"))

    @property
    def version(self) -> str:
        """Get the API version."""
        return str(self.info.get("version", "1.0.0"))

    @property
    def servers(self) -> list[dict[str, Any]]:
        """Get the servers list."""
        return list(self._spec.get("servers", []))

    @property
    def paths(self) -> dict[str, dict[str, Any]]:
        """Get the paths object."""
        return dict(self._spec.get("paths", {}))

    @property
    def components(self) -> dict[str, Any]:
        """Get the components object."""
        return dict(self._spec.get("components", {}))

    @property
    def schemas(self) -> dict[str, dict[str, Any]]:
        """Get all schemas from components."""
        return dict(self.components.get("schemas", {}))

    @property
    def security_schemes(self) -> dict[str, dict[str, Any]]:
        """Get security schemes from components."""
        return dict(self.components.get("securitySchemes", {}))

    @property
    def security(self) -> list[dict[str, list[str]]]:
        """Get top-level security requirements."""
        return list(self._spec.get("security", []))

    def resolve_ref(self, ref: str) -> dict[str, Any]:
        """Resolve a $ref pointer to its target."""
        if not ref.startswith("#/"):
            raise ValueError(f"Only local refs supported: {ref}")

        parts = ref[2:].split("/")
        current: dict[str, Any] | list[Any] | str | int | float | bool | None = (
            self._spec
        )
        for part in parts:
            part = part.replace("~1", "/").replace("~0", "~")
            if isinstance(current, dict):
                current = current.get(part, {})
            else:
                return {}
        return dict(current) if isinstance(current, dict) else {}

    def get_operations(self) -> list[tuple[str, str, dict[str, Any]]]:
        """Get all operations as (path, method, operation) tuples."""
        operations: list[tuple[str, str, dict[str, Any]]] = []
        # Use a list to ensure consistent ordering
        http_methods = [
            "get",
            "post",
            "put",
            "delete",
            "patch",
            "head",
            "options",
            "trace",
        ]

        for path, path_item in self.paths.items():
            for method in http_methods:
                if method in path_item:
                    operations.append((path, method, path_item[method]))

        return operations



"""Senzo: Generate typed Python API clients from OpenAPI specs."""

from senzo.generator import generate_tree, write_package
from senzo.settings import EnumStyle, GeneratorConfig, PaginationConfig, TagStyle

__all__ = [
    "generate_tree",
    "write_package",
    "EnumStyle",
    "GeneratorConfig",
    "PaginationConfig",
    "TagStyle",
]



"""Backend implementations for code generation."""

from senzo.backends.base import (
    DataclassBackend,
    HttpBackend,
    FieldDefinition,
    TypeMapping,
)
from senzo.backends.dataclass.msgspec import MsgspecBackend
from senzo.backends.dataclass.pydantic import PydanticBackend
from senzo.backends.http.httpx import HttpxBackend
from senzo.settings import EnumStyle

# Backwards compatibility alias
EnumStyleBackend = EnumStyle

__all__ = [
    "DataclassBackend",
    "EnumStyle",
    "EnumStyleBackend",  # Deprecated alias for EnumStyle
    "HttpBackend",
    "FieldDefinition",
    "TypeMapping",
    "MsgspecBackend",
    "PydanticBackend",
    "HttpxBackend",
]
